# See "Writing benchmarks" in the asv docs for more information.
# https://asv.readthedocs.io/en/latest/writing_benchmarks.html
# or the napari documentation on benchmarking
# https://github.com/napari/napari/blob/main/docs/BENCHMARKS.md
import itertools
from collections.abc import Callable
from contextlib import suppress
from enum import StrEnum, auto
from functools import cache, wraps
from typing import Any

import numpy as np

from napari.layers import Shapes
from napari.layers.shapes._shapes_constants import shape_classes
from napari.layers.shapes._shapes_models import Path, Polygon
from napari.settings import get_settings
from napari.utils._test_utils import read_only_mouse_event
from napari.utils.interactions import (
    mouse_move_callbacks,
    mouse_press_callbacks,
    mouse_release_callbacks,
)

try:
    from .utils import Skip
except ImportError:
    from napari.benchmarks.utils import Skip

try:
    from napari.utils.triangulation_backend import TriangulationBackend
except ImportError:

    class TriangulationBackend(StrEnum):
        partsegcore = (
            auto()
        )  # data preprocessing using PartSegCore (https://partseg.github.io)
        bermuda = auto()  # data preprocessing using bermuda (https://github.com/napari/bermuda)
        triangle = auto()  # data preprocessing using numba, edge triangulation using numba, triangulation using triangle
        numba = auto()  # data preprocessing and edge triangulation using numba, triangulation using vispy
        pure_python = auto()  # data preprocessing, edge triangulation, and triangulation using python and vispy

        def __str__(self):
            return self.name

        def __repr__(self):
            return self.name


backends = list(TriangulationBackend)

backend_list_complex = [
    TriangulationBackend.partsegcore,
    TriangulationBackend.bermuda,
    TriangulationBackend.triangle,
    TriangulationBackend.numba,
]
# pure python backend is too slow for large datasets
# triangle backend requires deduplication of vertices in #6654


class _BackendSelection:
    """Provides a representation for a preferred backend if accelerated triangulation is unavailable."""

    triangulate: Callable | None  # a triangulate function
    prev_settings_old: bool  # save the state of NapariSettings.experimental.compiled_triangulation,
    prev_settings_new: Any  # save the state of NapariSettings.experimental.triangulation_backend,
    prev_numba: dict[
        str, Callable
    ]  # Maps numba function names to _accelerated_triangulate_dispatch functions.

    def _disable_numba(self):
        """Replace numba jit functions with pure python functions."""
        try:
            from napari.layers.shapes import (
                _accelerated_triangulate_dispatch as _triangle_dispatch,
            )
        except ImportError:
            return

        if not hasattr(_triangle_dispatch, '__all__'):
            # new dispatch do not require this
            return

        for name in _triangle_dispatch.__all__:
            # This part of the code assumes that all pure python functions
            # from _accelerated_triangulate_dispatch that are used when import from
            # _accelerated_triangulate module fails because of missing numba
            # are named as the function that they replace with added _py suffix.
            # So we could write generic code to replace all of them.
            if hasattr(_triangle_dispatch, f'{name}_py'):
                self.prev_numba[name] = getattr(_triangle_dispatch, name)
                setattr(
                    _triangle_dispatch,
                    name,
                    getattr(_triangle_dispatch, f'{name}_py'),
                )

    def _set_settings_old(self, triangulation_backend: TriangulationBackend):
        with suppress(AttributeError):
            self.prev_settings = get_settings().experimental.backend_type
            get_settings().experimental.backend_type = (
                triangulation_backend == TriangulationBackend.partsegcore
            )
            get_settings().experimental.compiled_triangulation = (
                triangulation_backend
                in {
                    TriangulationBackend.partsegcore,
                    TriangulationBackend.bermuda,
                }
            )

    def _set_settings_new(self, triangulation_backend: TriangulationBackend):
        self.prev_settings_new = (
            get_settings().experimental.triangulation_backend
        )
        get_settings().experimental.triangulation_backend = (
            triangulation_backend
        )

    def select_backend(self, triangulation_backend: TriangulationBackend):
        """Select a desired backend for triangulation."""
        self.prev_numba = {}
        if hasattr(get_settings().experimental, 'triangulation_backend'):
            self._set_settings_new(triangulation_backend)
        else:
            self._set_settings_old(triangulation_backend)

        from napari.layers.shapes import _shapes_utils

        self.triangulate = _shapes_utils.triangulate
        if triangulation_backend != TriangulationBackend.triangle:
            # Disable the triangle backend by overriding the function
            _shapes_utils.triangulate = None
        if triangulation_backend == TriangulationBackend.pure_python:
            self._disable_numba()
        else:
            self.warmup_numba()

    @staticmethod
    def warmup_numba() -> None:
        """Warmup numba cache to avoid the first call being slow."""
        try:
            from napari.layers.shapes._accelerated_triangulate_dispatch import (
                warmup_numba_cache,
            )
        except ImportError:
            return
        warmup_numba_cache()

    def revert_backend(self):
        """Restore changes made by select_backend function. Call in teardown step."""
        if hasattr(get_settings().experimental, 'triangulation_backend'):
            get_settings().experimental.triangulation_backend = (
                self.prev_settings_new
            )
        else:
            with suppress(AttributeError):
                get_settings().experimental.compiled_triangulation = (
                    self.prev_settings
                )

        from napari.layers.shapes import _shapes_utils

        _shapes_utils.triangulate = self.triangulate
        with suppress(ImportError):
            from napari.layers.shapes import (
                _accelerated_triangulate_dispatch as atd,
            )

            for name, func in self.prev_numba.items():
                setattr(atd, name, func)

    def teardown(self, *_):
        self.revert_backend()


def skip_above_100(n_shapes, *_):
    return n_shapes > 100


class Shapes2DSuite(_BackendSelection):
    """Benchmarks for the Shapes layer with 2D data"""

    data: list[np.ndarray]
    layer: Shapes

    params = [tuple(2**i for i in range(4, 9)), backends]
    params_names = ['n_shapes', 'backend']

    skip_params = Skip(if_in_pr=lambda n_shapes, backend: n_shapes > 2**5)

    def setup(self, n_shapes, backend):
        self.select_backend(backend)
        rng = np.random.default_rng(0)
        self.data = [50 * rng.random((6, 2)) for _ in range(n_shapes)]
        self.layer = Shapes(self.data, shape_type='polygon')
        self.layer.selected_data = list(range(n_shapes))

    def time_create_layer(self, *_):
        """Time to create an image layer."""
        Shapes(self.data, shape_type='polygon')

    def time_refresh(self, *_):
        """Time to refresh view."""
        self.layer.refresh()

    def time_set_view_slice(self, *_):
        """Time to set view slice."""
        if hasattr(self.layer, '_slicing_state'):
            self.layer._slicing_state._set_view_slice()
        else:
            # before https://github.com/napari/napari/pull/8254
            self.layer._set_view_slice()

    def time_update_thumbnail(self, *_):
        """Time to update thumbnail."""
        self.layer._update_thumbnail()

    def time_get_value(self, *_):
        """Time to get current value."""
        for i in range(100):
            self.layer.get_value((i,) * 2)

    def mem_layer(self, *_):
        """Memory used by layer."""
        return self.layer

    def mem_data(self, *_):
        """Memory used by raw data."""
        return self.data

    def time_edit_shape(self, *_):
        """Time to edit a shape."""
        # Simulate editing the first shape
        self.layer._data_view.edit(
            0, np.array([[10, 10], [20, 20], [30, 30]]), new_type='polygon'
        )


class Shape2DEditSuite:
    params = [tuple(2**i for i in range(4, 9)), ('single', 'all')]
    params_names = ['n_shapes', 'selection']

    def setup(self, n_shapes, selection):
        rng = np.random.default_rng(0)
        self.data = [50 * rng.random((6, 2)) for _ in range(n_shapes)]
        self.layer = Shapes(self.data, shape_type='polygon')
        if selection == 'all':
            self.layer.selected_data = list(range(n_shapes))
        else:
            self.layer.selected_data = [2]

    def time_set_edge_width(self, *_):
        self.layer.current_edge_width = 10

    def time_set_edge_color(self, *_):
        self.layer.current_edge_color = 'red'  # RGBA red

    def time_set_face_color(self, *_):
        self.layer.current_face_color = 'red'


class Shapes3DSuite:
    """Benchmarks for the Shapes layer with 3D data."""

    data: list[np.ndarray]
    layer: Shapes

    params = [2**i for i in range(4, 9)]
    skip_params = Skip(if_in_pr=lambda n_shapes: n_shapes > 2**5)

    def setup(self, n):
        rng = np.random.default_rng(0)
        self.data = [50 * rng.random((6, 3)) for _ in range(n)]
        self.layer = Shapes(self.data, shape_type='polygon')

    def time_create_layer(self, _n):
        """Time to create a layer."""
        Shapes(self.data, shape_type='polygon')

    def time_refresh(self, _n):
        """Time to refresh view."""
        self.layer.refresh()

    def time_set_view_slice(self, _n):
        """Time to set view slice."""
        if hasattr(self.layer, '_slicing_state'):
            self.layer._slicing_state._set_view_slice()
        else:
            # before https://github.com/napari/napari/pull/8254
            self.layer._set_view_slice()

    def time_update_thumbnail(self, _n):
        """Time to update thumbnail."""
        self.layer._update_thumbnail()

    def time_get_value(self, _n):
        """Time to get current value."""
        self.layer.get_value((0,) * 3)

    def mem_layer(self, _n):
        """Memory used by layer."""
        return self.layer

    def mem_data(self, _n):
        """Memory used by raw data."""
        return self.data


class ShapesInteractionSuite:
    """Benchmarks for interacting with the Shapes layer with 2D data"""

    data: list[np.ndarray]
    layer: Shapes

    params = [2**i for i in range(4, 9)]

    def setup(self, n):
        rng = np.random.default_rng(0)
        self.data = [50 * rng.random((6, 2)) for _ in range(n)]
        self.layer = Shapes(self.data, shape_type='polygon')
        self.layer.mode = 'select'

        # initialize the position and select a shape
        position = tuple(np.mean(self.layer.data[0], axis=0))

        # create events
        click_event = read_only_mouse_event(
            type='mouse_press',
            is_dragging=False,
            modifiers=[],
            position=position,
        )
        # Simulate click
        mouse_press_callbacks(self.layer, click_event)

        release_event = read_only_mouse_event(
            type='mouse_release',
            is_dragging=False,
            modifiers=[],
            position=position,
        )

        # Simulate release
        mouse_release_callbacks(self.layer, release_event)

    def time_drag_shape(self, _n):
        """Time to process 5 shape drag events"""
        # initialize the position and select a shape
        position = tuple(np.mean(self.layer.data[0], axis=0))

        # create events
        click_event = read_only_mouse_event(
            type='mouse_press',
            is_dragging=False,
            modifiers=[],
            position=position,
        )

        # Simulate click
        mouse_press_callbacks(self.layer, click_event)

        # create events
        drag_event = read_only_mouse_event(
            type='mouse_press',
            is_dragging=True,
            modifiers=[],
            position=position,
        )

        # start drag event
        mouse_move_callbacks(self.layer, drag_event)

        # simulate 5 drag events
        for _ in range(5):
            position = tuple(np.add(position, [10, 5]))
            drag_event = read_only_mouse_event(
                type='mouse_press',
                is_dragging=True,
                modifiers=[],
                position=position,
            )

            # Simulate move, click, and release
            mouse_move_callbacks(self.layer, drag_event)

        release_event = read_only_mouse_event(
            type='mouse_release',
            is_dragging=False,
            modifiers=[],
            position=position,
        )

        # Simulate release
        mouse_release_callbacks(self.layer, release_event)

    time_drag_shape.param_names = ['n_shapes']

    def time_select_shape(self, _n):
        """Time to process shape selection events"""
        position = tuple(np.mean(self.layer.data[1], axis=0))

        # create events
        click_event = read_only_mouse_event(
            type='mouse_press',
            is_dragging=False,
            modifiers=[],
            position=position,
        )
        # Simulate click
        mouse_press_callbacks(self.layer, click_event)

        release_event = read_only_mouse_event(
            type='mouse_release',
            is_dragging=False,
            modifiers=[],
            position=position,
        )

        # Simulate release
        mouse_release_callbacks(self.layer, release_event)

    time_select_shape.param_names = ['n_shapes']


def get_shape_type(func):
    @wraps(func)
    def wrap(self, *args):
        shape_type_pos = self.param_names.index('shape_type')
        return func(self, args[shape_type_pos])

    return wrap


class _ShapeTriangulationBase(_BackendSelection):
    data: list[np.ndarray]

    @get_shape_type
    def time_create_layer(self, shape_type):
        """Time to create a Shapes layer when passed data and a shape type."""
        Shapes(self.data, shape_type=shape_type)


class _ShapeTriangulationBaseShapeCount(_ShapeTriangulationBase):
    param_names = [
        'n_shapes',
        'n_points',
        'shape_type',
        'triangulation_backend',
    ]
    params = [
        (
            100,
            5_000,
        ),
        (8, 32, 128),
        ('path', 'polygon'),
        backend_list_complex,
    ]

    skip_params = Skip(if_in_pr=skip_above_100)


class ShapeTriangulationNonConvexSuite(_ShapeTriangulationBaseShapeCount):
    skip_params = Skip(
        # skip a case when vispy triangulation backend fails
        always=lambda n_shapes,
        n_points,
        shape_type,
        triangulation_backend: n_points == 128
        and shape_type == 'polygon'
        and triangulation_backend != TriangulationBackend.triangle,
        if_in_pr=skip_above_100,
        # too slow (40 sec)
        if_on_ci=lambda n_shapes,
        n_points,
        shape_type,
        triangulation_backend: (
            n_shapes == 5000
            and n_points == 32
            and shape_type == 'polygon'
            and triangulation_backend
            in {
                TriangulationBackend.pure_python and TriangulationBackend.numba
            }
        ),
    )

    def setup(self, n_shapes, n_points, _shape_type, triangulation_backend):
        self.data = non_convex_no_self_intersection_polygons(
            n_shapes, n_points
        )
        self.select_backend(triangulation_backend)


class ShapeTriangulationConvexSuite(_ShapeTriangulationBaseShapeCount):
    def setup(self, n_shapes, n_points, _shape_type, triangulation_backend):
        self.data = convex_polygons(n_shapes, n_points)
        self.select_backend(triangulation_backend)


class ShapeTriangulationIntersectionSuite(_ShapeTriangulationBaseShapeCount):
    params = [
        (
            100,
            5_000,
        ),
        (7, 9, 15, 33),
        ('path', 'polygon'),
        backend_list_complex,
    ]
    skip_params = Skip(
        if_on_ci=lambda n_shapes,
        n_points,
        shape_type,
        triangulation_backend: triangulation_backend
        != TriangulationBackend.numba
        and n_shapes > 100,
        if_in_pr=skip_above_100,
    )

    def setup(self, n_shapes, n_points, _shape_type, triangulation_backend):
        self.data = self_intersecting_polygons(n_shapes, n_points)
        self.select_backend(triangulation_backend)


class ShapeTriangulationStarIntersectionSuite(
    _ShapeTriangulationBaseShapeCount
):
    params = [
        (
            100,
            5_000,
        ),
        (7, 9, 15, 33),
        ('path', 'polygon'),
        backend_list_complex,
    ]
    skip_params = Skip(
        always=lambda n_shapes,
        n_points,
        shape_type,
        triangulation_backend: n_shapes == 5000
        and n_points in {15, 33}
        and shape_type == 'polygon',
        if_on_ci=lambda n_shapes,
        n_points,
        shape_type,
        triangulation_backend: (
            (
                # 7 and 9 points are too slow to run
                n_shapes == 5000 and shape_type == 'polygon'
            )
            or (
                (n_shapes == 100 and n_points == 33)
                and shape_type == 'polygon'
                and triangulation_backend != TriangulationBackend.triangle
            )
        ),
        if_in_pr=skip_above_100,
    )

    def setup(self, n_shapes, n_points, _shape_type, triangulation_backend):
        self.data = self_intersecting_stars_polygons(n_shapes, n_points)
        self.select_backend(triangulation_backend)


def skip_triangle_backend(
    _n_shapes, _n_points, _shape_type, triangulation_backend
):
    """This is a helper function to skip triangle testing in polygon with holes case if
    the bugfix is not available.

    It may be removed after release on napari 0.6.0.
    """
    from napari.layers.shapes import _shapes_utils

    if hasattr(_shapes_utils, '_cull_triangles_not_in_poly'):
        return False
    return triangulation_backend == TriangulationBackend.triangle


class ShapeTriangulationHoleSuite(_ShapeTriangulationBaseShapeCount):
    params = [
        (
            100,
            1_000,
        ),
        (12, 48),
        ('path', 'polygon'),
        backend_list_complex,
    ]
    skip_params = Skip(
        always=skip_triangle_backend,
        if_in_pr=skip_above_100,
        if_on_ci=lambda n_shapes,
        n_points,
        shape_type,
        triangulation_backend: n_shapes > 100
        and triangulation_backend == TriangulationBackend.numba
        and shape_type == 'polygon',
    )

    def setup(self, n_shapes, n_points, _shape_type, triangulation_backend):
        self.data = polygons_with_hole(n_shapes, n_points)
        self.select_backend(triangulation_backend)


class ShapeTriangulationHolesSuite(_ShapeTriangulationBaseShapeCount):
    params = [
        (
            100,
            1_000,
        ),
        (24, 48),
        ('path', 'polygon'),
        backend_list_complex,
    ]
    skip_params = Skip(
        always=skip_triangle_backend,
        if_in_pr=skip_above_100,
        if_on_ci=lambda n_shapes,
        n_points,
        shape_type,
        triangulation_backend: n_shapes > 100
        and triangulation_backend == TriangulationBackend.numba
        and shape_type == 'polygon',
    )

    def setup(self, n_shapes, n_points, _shape_type, triangulation_backend):
        self.data = polygons_with_hole(n_shapes, n_points)
        self.select_backend(triangulation_backend)


class ShapeTriangulationMixed(_ShapeTriangulationBase):
    param_names = ['n_shapes', 'shape_type', 'triangulation_backend']
    params = [
        (
            100,
            3_000,
        ),
        ('path', 'polygon'),
        backend_list_complex,
    ]

    # the case of 128 points crashes the benchmark on call of PolygonData(vertices=data).triangulate()
    skip_params = Skip(if_in_pr=skip_above_100)

    def setup(self, n_shapes, _shape_type, triangulation_backend):
        part_size = int(n_shapes / 10)
        self.data = list(
            itertools.chain(
                convex_polygons(part_size, 4),
                convex_polygons(part_size * 2, 5),
                convex_polygons(part_size * 2, 7),
                convex_polygons(part_size, 60),
                non_convex_no_self_intersection_polygons(part_size, 10),
                non_convex_no_self_intersection_polygons(part_size, 60),
            )
        )
        self.select_backend(triangulation_backend)

    @get_shape_type
    def time_create_shapes(self, shape_type):
        """Time to create a layer."""
        cls = shape_classes[shape_type]
        for data in self.data:
            cls(data)


class MeshTriangulationSuite(_BackendSelection):
    data: list[Polygon | Path]

    param_names = ['shape_type', 'triangulation_backend']
    params = [('path', 'polygon'), backend_list_complex]

    def setup(self, shape_type, triangulation_backend):
        self.select_backend(triangulation_backend)
        part_size = 10
        self.data = [
            shape_classes[shape_type](x)
            for x in itertools.chain(
                convex_polygons(part_size, 4),
                convex_polygons(part_size * 2, 5),
                convex_polygons(part_size * 2, 12),
                convex_polygons(part_size, 60),
                non_convex_no_self_intersection_polygons(part_size, 10),
                non_convex_no_self_intersection_polygons(part_size, 60),
                self_intersecting_stars_polygons(part_size, 11),
                self_intersecting_polygons(part_size, 15),
            )
        ]

    def time_set_meshes(self, shape_type, _triangulation_backend):
        for shape in self.data:
            shape._set_meshes(
                shape.data,
                face=(shape_type == 'polygon'),
                closed=True,
                edge=True,
            )


@cache
def non_convex_no_self_intersection_polygons(
    n_shapes=5_000, n_points=32
) -> list[np.ndarray]:
    """
    Create a set of non-convex coordinates

    Parameters
    ----------
    n_shapes : int
        Number of shapes to create
    n_points : int
        Number of virtex of each shape
    """
    rng = np.random.default_rng(0)
    radius = 1000
    center = rng.uniform(1500, 4500, (n_shapes, 2))
    phi = np.linspace(0, 2 * np.pi, n_points)
    rays = np.stack([np.sin(phi), np.cos(phi)], axis=1)
    rays = rays.reshape((1, -1, 2))
    rays = rays * rng.uniform(0.9, 1.1, (n_shapes, n_points, 2))
    center = center.reshape((-1, 1, 2))
    return list(center + radius * rays)


@cache
def self_intersecting_stars_polygons(
    n_shapes=5_000, n_points=31
) -> list[np.ndarray]:
    """
    Create a set of non-convex coordinates

    Parameters
    ----------
    n_shapes : int
        Number of shapes to create
    n_points : int
        Number of virtex of each shape
    """
    assert n_points % 2 == 1
    rng = np.random.default_rng(0)
    radius = 5000
    center = rng.uniform(5000, 15000, (n_shapes, 2))
    shift = np.floor(n_points / 2) + 1
    phi = np.linspace(0, 2 * np.pi, n_points + 1) * shift
    rays = np.stack([np.sin(phi), np.cos(phi)], axis=1)
    rays = rays.reshape((1, -1, 2))
    rays = rays * rng.uniform(0.9, 1.1, (n_shapes, n_points + 1, 2))
    center = center.reshape((-1, 1, 2))
    return list(center + radius * rays)


@cache
def self_intersecting_polygons(
    n_shapes=5_000, n_points=31
) -> list[np.ndarray]:
    """
    Create a set of non-convex coordinates

    Parameters
    ----------
    n_shapes : int
        Number of shapes to create
    n_points : int
        Number of virtex of each shape
    """
    assert n_points % 2 == 1
    rng = np.random.default_rng(0)
    radius = 5000
    center = rng.uniform(5000, 15000, (n_shapes, 2))
    phi = np.linspace(0, 2 * np.pi, n_points + 1) * 2
    rays = np.stack([np.sin(phi), np.cos(phi)], axis=1)
    rays = rays.reshape((1, -1, 2))
    rays = rays * rng.uniform(0.9, 1.1, (n_shapes, n_points + 1, 2))
    center = center.reshape((-1, 1, 2))
    return list(center + radius * rays)


@cache
def convex_polygons(n_shapes=5_000, n_points=32) -> list[np.ndarray]:
    """
    Create a set of convex coordinates

    Parameters
    ----------
    n_shapes : int
        Number of shapes to create
    n_points : int
        Number of virtex of each shape
    """
    rng = np.random.default_rng(0)
    radius = 500 / np.sqrt(n_shapes)
    center = rng.uniform(500, 1500, (n_shapes, 2))
    phi = np.linspace(0, 2 * np.pi, n_points, endpoint=False)
    rays = np.stack([np.sin(phi), np.cos(phi)], axis=1)
    rays = rays.reshape((1, -1, 2))
    center = center.reshape((-1, 1, 2))
    return list(center + radius * rays)


@cache
def polygons_with_hole(n_shapes=5_000, n_points=32) -> list[np.ndarray]:
    """
    Create a set of polygon with hole

    Parameters
    ----------
    n_shapes : int
        Number of shapes to create
    n_points : int
        Number of virtex of each shape
    """
    rng = np.random.default_rng(0)
    assert n_points > 7, (
        'n_points should be greater than 7 to generate a polygon with holes'
    )
    radius = 500
    outer_points_num = n_points // 2 + 1
    inner_points_num = n_points - outer_points_num + 2

    center = rng.uniform(500, 1500, (n_shapes, 2))
    phi1 = np.linspace(0, 2 * np.pi, outer_points_num, endpoint=True)
    phi1[-1] = 0
    phi2 = np.linspace(0, 2 * np.pi, inner_points_num, endpoint=True)
    phi2[-1] = 0
    rays1 = np.stack([np.sin(phi1), np.cos(phi1)], axis=1) * radius
    rays2 = np.stack([np.sin(phi2), np.cos(phi2)], axis=1) * radius // 2
    rays = np.concatenate([rays1, rays2]).reshape((1, -1, 2))
    center = center.reshape((-1, 1, 2))
    return list(center + rays)


@cache
def polygons_with_holes(n_shapes=5_000, n_points=32) -> list[np.ndarray]:
    rng = np.random.default_rng(0)
    assert n_points > 20, (
        'n_points should be greater than 7 to generate a polygon with holes'
    )
    radius = 500
    inner_points_num = 4
    outer_points_num = n_points - 4 * inner_points_num

    rectangle = np.array([[1, 0], [0, 1], [-1, 0], [0, -1], [1, 0]]) * 50

    center = rng.uniform(500, 1500, (n_shapes, 2))
    phi = np.linspace(0, 2 * np.pi, outer_points_num, endpoint=True)
    phi[-1] = 0
    rays = np.stack([np.sin(phi), np.cos(phi)], axis=1) * radius
    steep = len(phi) // 4

    points = np.concatenate(
        [
            rays[:steep],
            rectangle + (250, 0),
            rays[steep - 1 : 2 * steep],
            rectangle + (0, -250),
            rays[2 * steep - 1 : 3 * steep],
            rectangle + (-250, 0),
            rays[3 * steep - 1 :],
            rectangle + (0, 250),
        ]
    ).reshape((1, -1, 2))

    center = center.reshape((-1, 1, 2))
    return list(center + points)


if __name__ == '__main__':
    from utils import run_benchmark

    run_benchmark()



"""
Export regions of interest (ROIs) to png
==========================================

Display multiple layer types one of them representing ROIs, add scale bar,
and export unscaled and scaled rois. The scale bar will always be within
the screenshot of the ROI.

Currently, `export_rois` does not support the 3D view or any other shape than 2D rectangles.

In the final grid state shown below, from left to right, top to bottom, the first 4 images display the
scaled roi screenshots with scale bar and the last 4 display the unscaled roi screenshots with scale bar.

.. tags:: visualization-advanced
"""

import numpy as np
from skimage import data

import napari

# Create a napari viewer with multiple layer types and add a scale bar.
# One of the polygon shapes exists outside the image extent, which is
# useful in displaying how figure export handles the extent of all layers.

viewer = napari.Viewer()

# add a 2D image layer
img_layer = viewer.add_image(data.lily(), name='lily', channel_axis=2)

# Rectangular shapes encapsulating rois for the lily data
rois = [np.array([[149.38401138,  61.80004348],
        [149.38401138, 281.95358131],
        [364.2538643 , 281.95358131],
        [364.2538643 ,  61.80004348]]),
 np.array([[316.70070013, 346.23841435],
        [316.70070013, 660.61766637],
        [673.34943141, 660.61766637],
        [673.34943141, 346.23841435]]),
 np.array([[579.12371722,  16.88872176],
        [579.12371722, 176.27988315],
        [768.45575975, 176.27988315],
        [768.45575975,  16.88872176]]),
 np.array([[ 43.42954911, 445.29831816],
        [ 43.42954911, 871.04161258],
        [285.58890617, 871.04161258],
        [285.58890617, 445.29831816]])]

# Add lily rois to the viewer for visualization purposes, not required for exporting roi screenshots.
roi_layer = viewer.add_shapes(
    rois, # in case of a shapes layer containing rectangular rois, pass on layer.data directly.
    edge_width=10,
    edge_color='green', # Optionally, set to [(0, 0, 0, 0)] * 4 to prevent edge color from showing in screenshots.
    face_color=[(0, 0, 0, 0)] * 4, # We do not want the face color to show up in the screenshots
    shape_type='rectangle',
    name='rois',
)

# add scale_bar with background box
viewer.scale_bar.visible = True
viewer.scale_bar.box = True
# viewer.scale_bar.length = 150  # prevent dynamic adjustment of scale bar length

# Take screenshots of the rois.
screenshot_rois = viewer.export_rois(rois)
# Optionally, save the exported rois in a directory of choice with name `roi_n.png` where n is the index of the roi:
# viewer.export_rois(rois, paths='home/data/exported_rois')
# Optionally, save the exported rois while specifying the location for each roi to be stored:
# viewer.export_rois(rois, paths=['first_roi.png', 'second_roi.png', 'third_roi.png', 'fourth_roi.png'])

# Also take scaled roi screenshots.
screenshot_rois_scaled = viewer.export_rois(rois, scale=2
                                     )
viewer.layers.select_all()
viewer.layers.toggle_selected_visibility()
for index, roi in enumerate(screenshot_rois):
    viewer.add_image(roi, name=f'roi_{index}_unscaled')

for index, roi in enumerate(screenshot_rois_scaled):
    viewer.add_image(roi, name=f'roi_{index}_scaled')

viewer.grid.enabled = True
viewer.grid.shape = (3, 3)


if __name__ == '__main__':
    napari.run()



from importlib.metadata import version

from packaging.version import parse as parse_version

from napari.utils import _appdirs


def test_maybe_migrate_uvx_settings_no_uv(monkeypatch):
    monkeypatch.setattr(_appdirs, 'environment_marker', 'aaa')
    assert not _appdirs._maybe_migrate_uvx_settings()


def test_maybe_migrate_uvx_settings_no_config(monkeypatch, tmp_path):
    monkeypatch.setattr(_appdirs, 'environment_marker', 'uvx/aaa')
    monkeypatch.setattr(
        _appdirs, 'user_config_dir', lambda: str(tmp_path / 'config' / '0.1.0')
    )
    assert not _appdirs._maybe_migrate_uvx_settings()


def test_maybe_migrate_uvx_settings_config_exists(monkeypatch, tmp_path):
    monkeypatch.setattr(_appdirs, 'environment_marker', 'uvx/aaa')
    config_path = tmp_path / 'config' / '0.1.0'
    config_path.mkdir(parents=True)
    monkeypatch.setattr(_appdirs, 'user_config_dir', lambda: str(config_path))
    assert not _appdirs._maybe_migrate_uvx_settings()


def test__maybe_migrate_uvx_settings_migrate(monkeypatch, tmp_path):
    napari_version = parse_version(version('napari')).base_version
    monkeypatch.setattr(
        _appdirs, 'environment_marker', f'uvx/{napari_version}'
    )
    base_config_path = tmp_path / 'config'
    base_config_path.mkdir(parents=True)
    (base_config_path / '0.1.0').mkdir()
    (base_config_path / '0.1.0' / 'aa.txt').write_text('aa')
    (base_config_path / '0.2.0').mkdir()
    (base_config_path / '0.2.0' / 'bb.txt').write_text('bb')
    (base_config_path / '1000.2.0').mkdir()
    (base_config_path / '1000.2.0' / 'cc.txt').write_text('cc')

    monkeypatch.setattr(
        _appdirs,
        'user_config_dir',
        lambda: str(base_config_path / napari_version),
    )
    assert _appdirs._maybe_migrate_uvx_settings()
    assert (base_config_path / napari_version).exists()
    assert (base_config_path / napari_version / 'bb.txt').exists()



from app_model.types import (
    Action,
    SubmenuItem,
)

from napari._app_model.constants import MenuGroup, MenuId
from napari.components import LayerList, ViewerModel
from napari.layers import Layer, Points, Shapes
from napari.utils.translations import trans

FILE_SUBMENUS = [
    (
        MenuId.MENUBAR_FILE,
        SubmenuItem(
            submenu=MenuId.FILE_NEW_LAYER,
            title=trans._('New Layer'),
            group=MenuGroup.NAVIGATION,
            order=0,
        ),
    ),
    (
        MenuId.MENUBAR_FILE,
        SubmenuItem(
            submenu=MenuId.FILE_OPEN_WITH_PLUGIN,
            title=trans._('Open with Plugin'),
            group=MenuGroup.OPEN,
            order=99,
        ),
    ),
    (
        MenuId.MENUBAR_FILE,
        SubmenuItem(
            submenu=MenuId.FILE_SAMPLES,
            title=trans._('Open Sample'),
            group=MenuGroup.OPEN,
            order=100,
        ),
    ),
    (
        MenuId.MENUBAR_FILE,
        SubmenuItem(
            submenu=MenuId.FILE_IO_UTILITIES,
            title=trans._('IO Utilities'),
            group=MenuGroup.UTIL,
            order=101,
        ),
    ),
    (
        MenuId.MENUBAR_FILE,
        SubmenuItem(
            submenu=MenuId.FILE_ACQUIRE,
            title=trans._('Acquire'),
            group=MenuGroup.UTIL,
            order=101,
        ),
    ),
]


def add_new_points(viewer: 'ViewerModel') -> None:
    if not viewer.layers.selection:
        viewer.add_points(  # type: ignore[attr-defined]
            ndim=max(viewer.dims.ndim, 2),
            scale=viewer.layers.extent.step,
        )
    else:
        extent = viewer.layers.get_extent(viewer.layers.selection)
        viewer.add_points(  # type: ignore[attr-defined]
            ndim=len(extent.step),
            scale=extent.step,
        )


def add_new_shapes(viewer: 'ViewerModel') -> None:
    if not viewer.layers.selection:
        viewer.add_shapes(  # type: ignore[attr-defined]
            ndim=max(viewer.dims.ndim, 2),
            scale=viewer.layers.extent.step,
        )
    else:
        extent = viewer.layers.get_extent(viewer.layers.selection)
        viewer.add_shapes(  # type: ignore[attr-defined]
            ndim=len(extent.step),
            scale=extent.step,
        )


def _create_single_layer(
    source_layer: Layer,
    layer_class: type[Points] | type[Shapes],
    layer_name: str,
) -> Points | Shapes:
    """Create a single Points or Shapes layer from the given source layer."""
    return layer_class(
        name=layer_name,
        ndim=source_layer.ndim,
        scale=source_layer.scale,
        translate=source_layer.translate,
        rotate=source_layer.rotate,
        shear=source_layer.shear,
        units=source_layer.units,
        affine=source_layer.affine.affine_matrix,
    )


def get_layer_name(base_name: str, existing_names: set[str]) -> str:
    """Generate a unique layer name based on the base name and existing names."""
    if base_name not in existing_names:
        return base_name
    index = 1
    while f'{base_name} {index}' in existing_names:
        index += 1
    return f'{base_name} {index}'


def new_labels(viewer: ViewerModel) -> None:
    viewer._new_labels()


def _new_layer_from_active(
    layer_list: LayerList, layer_class: type[Points] | type[Shapes]
) -> Points | Shapes:
    """Create a new layer from the given layer list."""
    if layer_list.selection.active is None:
        raise ValueError('No active layer to create new layer from.')
    source_layer = layer_list.selection.active
    new_layer_name = get_layer_name(
        f'{source_layer.name} - {layer_class.__name__.lower()}',
        existing_names={layer.name for layer in layer_list},
    )
    return _create_single_layer(source_layer, layer_class, new_layer_name)


def new_points(viewer: ViewerModel) -> None:
    if viewer.layers.selection.active is not None:
        viewer.add_layer(_new_layer_from_active(viewer.layers, Points))
    else:
        add_new_points(viewer)


def new_shapes(viewer: ViewerModel) -> None:
    if viewer.layers.selection.active is not None:
        viewer.add_layer(_new_layer_from_active(viewer.layers, Shapes))
    else:
        add_new_shapes(viewer)


FILE_ACTIONS: list[Action] = [
    Action(
        id='napari.window.file.new_layer.new_labels',
        title=trans._('Labels'),
        callback=new_labels,
        menus=[{'id': MenuId.FILE_NEW_LAYER, 'group': MenuGroup.NAVIGATION}],
    ),
    Action(
        id='napari.window.file.new_layer.new_points',
        title=trans._('Points'),
        callback=new_points,
        menus=[{'id': MenuId.FILE_NEW_LAYER, 'group': MenuGroup.NAVIGATION}],
    ),
    Action(
        id='napari.window.file.new_layer.new_shapes',
        title=trans._('Shapes'),
        callback=new_shapes,
        menus=[{'id': MenuId.FILE_NEW_LAYER, 'group': MenuGroup.NAVIGATION}],
    ),
]



# /// script
# requires-python = ">=3.12"
# dependencies = [
#     "requests>=2.32.5",
# ]
# ///
"""
This file contains the code to create a PR or update an existing one based on the state of the current branch.
"""

import logging
import os
import subprocess  # nosec
from contextlib import contextmanager
from os import chdir, environ, getcwd
from pathlib import Path

import requests
from check_updated_packages import get_changed_dependencies

REPO_DIR = Path(__file__).parent.parent / 'napari_repo'
# GitHub API base URL
BASE_URL = 'https://api.github.com'
DEFAULT_BRANCH_NAME_PREFIX = 'auto-update-dependencies'
DEFAULT_BRANCH_NAME = f'{DEFAULT_BRANCH_NAME_PREFIX}-main'


@contextmanager
def cd(path: Path):
    """
    Change directory to the given path and return to the previous one afterwards.
    """
    current_dir = getcwd()
    try:
        chdir(path)
        yield
    finally:
        chdir(current_dir)


def _setup_git_author():
    subprocess.run(
        ['git', 'config', '--global', 'user.name', 'napari-bot'], check=True
    )  # nosec
    subprocess.run(
        [
            'git',
            'config',
            '--global',
            'user.email',
            'napari-bot@users.noreply.github.com',
        ],
        check=True,
    )  # nosec


def create_commit(message: str, branch_name: str = ''):
    """
    Create a commit calling git.
    """
    with cd(REPO_DIR):
        if branch_name:
            subprocess.run(['git', 'checkout', '-B', branch_name], check=True)
        subprocess.run(['git', 'add', '-u'], check=True)  # nosec
        subprocess.run(['git', 'commit', '-m', message], check=True)  # nosec


def push(branch_name: str, update: bool = False):
    """
    Push the current branch to the remote.
    """
    try:
        with cd(REPO_DIR):
            logging.info('go to dir %s', REPO_DIR)
            if update:
                logging.info('Pushing to %s', branch_name)
                subprocess.run(
                    [
                        'git',
                        'push',
                        '--force',
                        '--set-upstream',
                        'napari-bot',
                        branch_name,
                    ],
                    check=True,
                    capture_output=True,
                )
            else:
                logging.info('Force pushing to %s', branch_name)
                subprocess.run(
                    [
                        'git',
                        'push',
                        '--force',
                        '--set-upstream',
                        'napari-bot',
                        branch_name,
                    ],
                    check=True,
                    capture_output=True,
                )  # nosec
    except subprocess.CalledProcessError as e:
        stdout = e.stdout.decode() if e.stdout else ''
        stderr = e.stderr.decode() if e.stderr else ''
        logging.exception(
            'Error pushing to branch %s:\n\nstdout: %s\n\nstderr: %s\n\nWith error %d',
            branch_name,
            stdout,
            stderr,
            e.returncode,
        )
        raise


def commit_message(branch_name) -> str:
    with cd(REPO_DIR):
        changed_direct = get_changed_dependencies(
            all_packages=False,
            base_branch=branch_name,
            python_version='3.11',
            src_dir=REPO_DIR,
        )
    if not changed_direct:
        return 'Update indirect dependencies'
    return 'Update ' + ', '.join(f'`{x}`' for x in changed_direct)


def long_description(branch_name: str) -> str:
    with cd(REPO_DIR):
        all_changed = get_changed_dependencies(
            all_packages=True,
            base_branch=branch_name,
            python_version='3.11',
            src_dir=REPO_DIR,
        )
    return 'Updated packages: ' + ', '.join(f'`{x}`' for x in all_changed)


def create_pr_with_push(branch_name: str, access_token: str, repo=''):
    """
    Create a PR.
    """
    new_branch_name = f'{DEFAULT_BRANCH_NAME_PREFIX}-{branch_name}'

    if not repo:
        repo = os.environ.get('GITHUB_REPOSITORY', 'napari/napari')
    with cd(REPO_DIR):
        subprocess.run(['git', 'checkout', '-B', new_branch_name], check=True)
    create_commit(commit_message(branch_name))
    push(new_branch_name)

    logging.info('Create PR for branch %s', new_branch_name)
    if pr_number := list_pr_for_branch(
        new_branch_name, access_token, repo=repo
    ):
        update_own_pr(pr_number, access_token, branch_name, repo)
    else:
        create_pr(
            base_branch=branch_name,
            new_branch=new_branch_name,
            access_token=access_token,
            repo=repo,
            source_user='napari-bot',
        )


def update_own_pr(pr_number: int, access_token: str, base_branch: str, repo):
    headers = {'Authorization': f'token {access_token}'}
    payload = {
        'title': commit_message(base_branch),
        'body': long_description(base_branch),
    }
    url = f'{BASE_URL}/repos/{repo}/pulls/{pr_number}'
    logging.info('Update PR with payload: %s in %s', str(payload), url)
    response = requests.post(url, headers=headers, json=payload)
    response.raise_for_status()

    url_labels = f'{BASE_URL}/repos/{repo}/issues/{pr_number}/labels'
    response = requests.get(url_labels, headers=headers)
    response.raise_for_status()

    remove_label_url = (
        f'{BASE_URL}/repos/{repo}/issues/{pr_number}/labels/ready%20to%20merge'
    )

    # following lines is to check if "ready to merge" label is added to PR,
    # if it is present, then remove it to point that PR was changed
    for label in response.json():
        if label['name'] == 'ready to merge':
            response = requests.delete(remove_label_url, headers=headers)
            response.raise_for_status()
            break


def list_pr_for_branch(
    branch_name: str,
    access_token: str,
    repo: str = 'napari/napari',
    user: str = 'napari-bot',
):
    """
    check if PR for branch exists
    """
    url = f'{BASE_URL}/repos/{repo}/pulls?state=open&head={user}:{branch_name}'
    response = requests.get(url)
    response.raise_for_status()
    if response.json():
        return response.json()[0]['number']
    return None


def create_pr(
    base_branch: str, new_branch: str, access_token: str, repo, source_user=''
):
    # Prepare the headers with the access token
    headers = {'Authorization': f'token {access_token}'}

    # publish the comment
    payload = {
        'title': commit_message(base_branch),
        'body': long_description(base_branch),
        'head': new_branch,
        'base': base_branch,
        'maintainer_can_modify': True,
    }
    if source_user:
        payload['head'] = f'{source_user}:{new_branch}'
    pull_request_url = f'{BASE_URL}/repos/{repo}/pulls'
    logging.info(
        'Create PR with payload: %s in %s', str(payload), pull_request_url
    )
    response = requests.post(pull_request_url, headers=headers, json=payload)
    response.raise_for_status()
    logging.info('PR created: %s', response.json()['html_url'])
    add_label(repo, response.json()['number'], 'maintenance', access_token)


def add_label(repo, pr_num, label, access_token):
    pull_request_url = f'{BASE_URL}/repos/{repo}/issues/{pr_num}/labels'
    headers = {'Authorization': f'token {access_token}'}
    payload = {'labels': [label]}

    logging.info('Add labels: %s in %s', str(payload), pull_request_url)
    response = requests.post(pull_request_url, headers=headers, json=payload)
    response.raise_for_status()
    logging.info('Labels added: %s', response.json())


def add_comment_to_pr(
    pull_request_number: int,
    message: str,
    repo='napari/napari',
):
    """
    Add a comment to an existing PR.
    """
    # Prepare the headers with the access token
    headers = {'Authorization': f'token {os.environ.get("GITHUB_TOKEN")}'}

    # publish the comment
    payload = {'body': message}
    comment_url = (
        f'{BASE_URL}/repos/{repo}/issues/{pull_request_number}/comments'
    )
    response = requests.post(comment_url, headers=headers, json=payload)
    response.raise_for_status()


def update_pr(branch_name: str):
    """
    Update an existing PR.
    """
    pr_number = get_pr_number()

    target_repo = os.environ.get('FULL_NAME')

    new_branch_name = f'auto-update-dependencies/{target_repo}/{branch_name}'

    if (
        target_repo == os.environ.get('GITHUB_REPOSITORY', 'napari/napari')
        and branch_name == DEFAULT_BRANCH_NAME
    ):
        new_branch_name = DEFAULT_BRANCH_NAME

    create_commit(commit_message(branch_name), branch_name=new_branch_name)
    comment_content = long_description(f'origin/{branch_name}')

    try:
        push(new_branch_name, update=branch_name != DEFAULT_BRANCH_NAME)
    except subprocess.CalledProcessError as e:
        if 'create or update workflow' in e.stderr.decode():
            logging.info('Workflow file changed. Skip PR create.')
            comment_content += (
                '\n\n This PR contains changes to the workflow file. '
            )
            comment_content += 'Please download the artifact and update the constraints files manually. '
            comment_content += f'Artifact: https://github.com/{os.environ.get("GITHUB_REPOSITORY", "napari/napari")}/actions/runs/{os.environ.get("GITHUB_RUN_ID")}'
        else:
            raise
    else:
        if new_branch_name != DEFAULT_BRANCH_NAME:
            comment_content += update_external_pr_comment(
                target_repo, branch_name, new_branch_name
            )

    add_comment_to_pr(
        pr_number,
        comment_content,
        repo=os.environ.get('GITHUB_REPOSITORY', 'napari/napari'),
    )
    logging.info('PR updated: %s', pr_number)


def update_external_pr_comment(
    target_repo: str, branch_name: str, new_branch_name: str
) -> str:
    comment = '\n\nThis workflow cannot automatically update your PR or create PR to your repository. '
    comment += 'But you could open such PR by clicking the link: '
    comment += f'https://github.com/{target_repo}/compare/{branch_name}...napari-bot:{new_branch_name}.'
    comment += '\n\n'
    comment += 'You could also get the updated files from the '
    comment += f'https://github.com/napari-bot/napari/tree/{new_branch_name}/resources/constraints. '
    comment += 'Or ask the maintainers to provide you the contents of the constraints artifact '
    comment += f'from the run https://github.com/{os.environ.get("GITHUB_REPOSITORY", "napari/napari")}/actions/runs/{os.environ.get("GITHUB_RUN_ID")}'
    return comment


def get_pr_number() -> int:
    """
    Get the PR number from the environment based on the PR_NUMBER variable.

    Returns
    -------
    pr number: int
    """
    pr_number = environ.get('PR_NUMBER')
    logging.info('PR_NUMBER: %s', pr_number)
    return int(pr_number)


def main():
    event_name = environ.get('GITHUB_EVENT_NAME')
    branch_name = environ.get('BRANCH')

    access_token = environ.get('GHA_TOKEN_MAIN_REPO')

    _setup_git_author()

    logging.basicConfig(level=logging.INFO)
    logging.info('Branch name: %s', branch_name)
    logging.info('Event name: %s', event_name)

    if event_name in {'schedule', 'workflow_dispatch'}:
        logging.info('Creating PR')
        create_pr_with_push(branch_name, access_token)
    elif event_name in {'issue_comment', 'workflow_call'}:
        logging.info('Updating PR')
        update_pr(branch_name)
    elif event_name == 'pull_request':
        logging.info(
            'Pull request run. We cannot add comment or create PR. Please download the artifact.'
        )
    else:
        raise ValueError(f'Unknown event name: {event_name}')


if __name__ == '__main__':
    main()



import json
import logging
import pathlib
from argparse import ArgumentParser

import matplotlib.pyplot as plt

logging.basicConfig(
    format='%(levelname)s : %(asctime)s : %(message)s',
    level=logging.INFO,
)

parser = ArgumentParser(
    description='Plot the durations of a callable measured by perfmon.',
)
parser.add_argument(
    'config',
    help='The name of the sub-directory that contains the perfmon traces (e.g. slicing)',
)
parser.add_argument(
    'callable',
    help='The name of the callable to plot excluding the module (e.g. QtDimSliderWidget._value_changed).',
)
parser.add_argument(
    '--output', default='latest', help='The name added to output traces file.'
)
args = parser.parse_args()

logging.info(
    'Running plot_callable.py with the following arguments.\n{args_}',
    extra={'args_': args},
)

perfmon_dir = pathlib.Path(__file__).parent.resolve(strict=True)

traces_path = perfmon_dir / args.config / f'traces-{args.output}.json'

with open(traces_path) as traces_file:
    traces = json.load(traces_file)

durations_ms = [
    trace['dur'] / 1000 for trace in traces if trace['name'] == args.callable
]

plt.violinplot(durations_ms, vert=False, showmeans=True, showmedians=True)

plt.title(f'{args.config} ({args.output}): {args.callable}')
plt.xlabel('Duration (ms)')
plt.yticks([])

plt.show()



"""
Pass colormaps
==============

Add named or unnamed vispy colormaps to existing layers.

.. tags:: visualization-basic
"""

import numpy as np
from skimage import data

import napari

histo = data.astronaut() / 255
rch, gch, bch = np.transpose(histo, (2, 0, 1))

v = napari.Viewer()

rlayer = v.add_image(
    rch, name='red channel', colormap='red', blending='additive'
)
glayer = v.add_image(
    gch, name='green channel', colormap='green', blending='additive'
)
blayer = v.add_image(
    bch, name='blue channel', colormap='blue', blending='additive'
)

if __name__ == '__main__':
    napari.run()



from typing import Any

from vispy.scene.visuals import Text as BaseText

from napari._vispy.utils.text import (
    get_text_width_height,
    register_napari_fonts,
)


class Text(BaseText):
    def __init__(
        self, *args: Any, face: str = 'AlataPlus', **kwargs: Any
    ) -> None:
        register_napari_fonts()
        super().__init__(*args, face=face, **kwargs)

    def get_width_height(self) -> tuple[float, float]:
        width, height = get_text_width_height(self)
        # width is not quite right for some reason... magic number here we go
        return width * self.dpi_ratio * 1.2, height * self.dpi_ratio

    @property
    def font_size(self) -> float:
        return self._font_size

    @font_size.setter
    def font_size(self, size: float) -> None:
        adjusted_font_size = size / self.dpi_ratio
        self._font_size = max(0.0, adjusted_font_size)
        self.update()

    @property
    def dpi_ratio(self) -> float:
        # adjust for dpi: 72 is the "base dpi" around which font sizes are defined
        # TODO: but for some reason 96 seems to give the correct ratio for me?
        return (self.transforms.dpi or 96) / 96



from datetime import date
from functools import lru_cache
from importlib import resources
from pathlib import Path


@lru_cache
def available_logos() -> list[str]:
    logo_dir = Path(resources.files('napari').joinpath('resources', 'logos'))  # type: ignore
    variants = ['auto']
    for logo in Path(logo_dir).glob('*-plain-light.svg'):
        variants.append(logo.stem.rsplit('-', 2)[0])
    return sorted(variants)


def _get_seasonal_logo(today: date | None = None, theme: str = 'dark') -> str:
    today = today or date.today()

    ranges = {
        'halloween': ((10, 25), (11, 2)),
        'christmas': ((12, 1), (1, 6)),
        'maythefourth': ((5, 1), (5, 10)),  # let's give it some leeway :P
    }

    theme_variants = {'maythefourth': {'dark': 'sith', 'light': 'jedi'}}

    for name, ((m1, d1), (m2, d2)) in ranges.items():  # noqa: B007
        start = date(today.year, m1, d1)
        end = date(today.year, m2, d2)

        if end < start:
            # rolls over to the next year
            if today >= start or today <= date(today.year, m2, d2):
                break
        else:
            if start <= today <= end:
                break

    else:
        name = 'gradient'

    if name in theme_variants and theme in theme_variants[name]:
        name = theme_variants[name][theme]

    return name


def get_logo_path(
    logo: str, template: str, theme: str, today: date | None = None
) -> Path:
    logo_dir = Path(resources.files('napari').joinpath('resources', 'logos'))  # type: ignore
    if logo not in available_logos():
        raise ValueError(
            f'logo must be one one {set(available_logos())}; got {logo}'
        )

    if template not in {'plain', 'padded'}:
        raise ValueError('template must be either "plain" or "padded"')

    # eventually we should actually use the dark/light "type" that the theme spec of npe2 allows,
    # which is currently unused in napari
    if theme not in {'dark', 'light'}:
        theme = 'dark'

    if logo == 'auto':
        logo = _get_seasonal_logo(today=today, theme=theme)

    return logo_dir / f'{logo}-{template}-{theme}.svg'



"""
Set colormaps
=============

Add named or unnamed vispy colormaps to existing layers.

.. tags:: visualization-basic
"""

import numpy as np
import vispy.color
from skimage import data

import napari

histo = data.astronaut() / 255
rch, gch, bch = np.transpose(histo, (2, 0, 1))
red = vispy.color.Colormap([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]])
green = vispy.color.Colormap([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
blue = vispy.color.Colormap([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]])

v = napari.Viewer()

rlayer = v.add_image(rch, name='red channel')
rlayer.blending = 'additive'
rlayer.colormap = 'red', red
glayer = v.add_image(gch, name='green channel')
glayer.blending = 'additive'
glayer.colormap = green  # this will appear as [unnamed colormap]
blayer = v.add_image(bch, name='blue channel')
blayer.blending = 'additive'
blayer.colormap = {'blue': blue}

if __name__ == '__main__':
    napari.run()



"""
Set theme
=========

Displays an image and sets the theme to 'light'.

.. tags:: gui
"""

from skimage import data

import napari

# create the viewer with an image
viewer = napari.Viewer()
layer = viewer.add_image(data.astronaut(), rgb=True, name='astronaut')

# set the theme to 'light'
viewer.theme = 'light'

if __name__ == '__main__':
    napari.run()



from unittest.mock import Mock

import numpy as np
import pytest

from napari.layers import Labels
from napari.layers.labels import _labels_key_bindings
from napari.layers.labels._labels_key_bindings import (
    decrease_label_id,
    increase_label_id,
    new_label,
    swap_selected_and_background_labels,
)


@pytest.fixture
def labels_data_4d():
    labels = np.zeros((5, 7, 8, 9), dtype=int)
    labels[1, 2:4, 4:6, 4:6] = 1
    labels[1, 3:5, 5:7, 6:8] = 2
    labels[2, 3:5, 5:7, 6:8] = 3
    return labels


def test_max_label(labels_data_4d):
    labels = Labels(labels_data_4d)
    new_label(labels)
    assert labels.selected_label == 4


def test_swap_background_label(labels_data_4d):
    labels = Labels(labels_data_4d)
    labels.selected_label = 10
    swap_selected_and_background_labels(labels)
    assert labels.selected_label == labels.colormap.background_value
    swap_selected_and_background_labels(labels)
    assert labels.selected_label == 10


def test_guard_for_out_of_range_selected_label(monkeypatch):
    show_warning_mock = Mock()
    monkeypatch.setattr(
        _labels_key_bindings, 'show_warning', show_warning_mock
    )
    labels = Labels(np.zeros((10, 10), dtype=np.uint8))

    labels.selected_label = 0
    decrease_label_id(labels)
    assert labels.selected_label == 0
    show_warning_mock.assert_called_once()
    show_warning_mock.call_args_list[0][0][0].startswith('The value -1')
    show_warning_mock.reset_mock()

    labels.selected_label = 255
    increase_label_id(labels)
    assert labels.selected_label == 255
    show_warning_mock.assert_called_once()
    show_warning_mock.call_args_list[0][0][0].startswith('The value 256')


def test_label_overflow(monkeypatch):
    show_warning_mock = Mock()
    monkeypatch.setattr(
        _labels_key_bindings, 'show_warning', show_warning_mock
    )
    data = np.zeros((10, 10), dtype=np.uint8)
    labels = Labels(data)
    new_label(labels)
    assert labels.selected_label == 1
    show_warning_mock.assert_not_called()

    data[0, 0] = 255
    new_label(labels)
    assert labels.selected_label == 1
    show_warning_mock.assert_called_once()
    show_warning_mock.call_args_list[0][0][0].startswith('The value 256')



"""
Displaying xarray data in napari
================================

This example shows how to view xarray datasets in napari, including scale
and translation information.

Currently, napari cannot display irregularly-sampled data, so the code
assumes that the data indices are regularly spaced. If your indices are
irregular, use `xarray.Dataset.interp` to create a regularly-spaced version
before displaying it in napari.

.. tags:: visualization-advanced, layers, xarray
"""
import numpy as np
import xarray as xr

import napari

# open the xarray global sea surface temperature (40MB) and North America
# air temperature (30MB) datasets
sst = xr.tutorial.open_dataset('ersstv5')
airtemp = xr.tutorial.open_dataset('air_temperature')


def get_scale_translate(dataset, array_name):
    """Get the translate/offset and scale parameters for an xarray dataset.

    This code assumes that the dataset is regularly spaced. You should
    interpolate your data if it is sampled at irregular spaces.

    Parameters
    ----------
    dataset : xr.Dataset
        The dataset containing the array to be displayed.
    array_name : str
        The name of the xarray DataArray within `dataset` to be displayed in
        napari.

    Returns
    -------
    param_dict : dict[str, list[float]]
        The scale and translate parameters computed from the xarray dimension
        indices.
    """
    array = getattr(dataset, array_name)
    if array is None:
        raise ValueError(f'{dataset} has no array with name {array_name}')
    dims = [getattr(dataset, dim) for dim in array.dims]
    translate = [float(d[0]) for d in dims]
    scale = [float(d[1] - d[0]) for d in dims]
    return {'scale': scale, 'translate': translate}


# Show the raw (not resampled) model data
viewer, sst_layer = napari.imshow(
        sst.sst,
        name='sea surface temp',
        **get_scale_translate(sst, 'sst'),
        colormap='magma',
        )
viewer.dims.axis_labels = sst.sst.dims

air_layer = viewer.add_image(
        airtemp.air,
        name='air temp NA',
        **get_scale_translate(airtemp, 'air'),
        colormap='viridis',
        blending='additive',
        contrast_limits=(-23 + 273, 32 + 273),  # data are in degrees Kelvin
        )

# set a time that overlaps both datasets
viewer.dims.set_point(0, np.datetime64('2013-03-10T18:00:00.000000000'))

# latitude goes from -90 (south, down) to 90 (north, up),
# so we make sure that the camera vertical axis points up.
viewer.camera.orientation2d = ('up', 'right')

# fill the frame
viewer.reset_view(margin=0)


if __name__ == '__main__':
    napari.run()



"""
Custom mouse functions
======================

Display one 4-D image layer using the ``add_image`` API

.. tags:: gui
"""

import numpy as np
from scipy import ndimage as ndi
from skimage import data
from skimage.morphology import binary_dilation, binary_erosion

import napari

np.random.seed(1)
viewer = napari.Viewer()
blobs = data.binary_blobs(length=128, volume_fraction=0.1, n_dim=2)
labeled = ndi.label(blobs)[0]
labels_layer = viewer.add_labels(labeled, name='blob ID')

@viewer.mouse_drag_callbacks.append
def get_event(viewer, event):
    print(event)

@viewer.mouse_drag_callbacks.append
def get_ndisplay(viewer, event):
    if 'Alt' in event.modifiers:
        print('viewer display ', viewer.dims.ndisplay)

@labels_layer.mouse_drag_callbacks.append
def get_connected_component_shape(layer, event):
    data_coordinates = layer.world_to_data(event.position)
    cords = np.round(data_coordinates).astype(int)
    val = layer.get_value(data_coordinates)
    if val is None:
        return
    if val != 0:
        data = layer.data
        binary = data == val
        if 'Shift' in event.modifiers:
            binary_new = binary_erosion(binary)
            data[binary] = 0
        else:
            binary_new = binary_dilation(binary)
        data[binary_new] = val
        size = np.sum(binary_new)
        layer.data = data
        msg = (
            f'clicked at {cords} on blob {val} which is now {size} pixels'
        )
    else:
        msg = f'clicked at {cords} on background which is ignored'
    print(msg)

# Handle click or drag events separately
@labels_layer.mouse_drag_callbacks.append
def click_drag(layer, event):
    print('mouse down')
    dragged = False
    yield
    # on move
    while event.type == 'mouse_move':
        print(event.position)
        dragged = True
        yield
    # on release
    if dragged:
        print('drag end')
    else:
        print('clicked!')

# Handle click or drag events separately
@labels_layer.mouse_double_click_callbacks.append
def on_second_click_of_double_click(layer, event):
    print('Second click of double_click', event.position)
    print('note that a click event was also triggered', event.type)


if __name__ == '__main__':
    napari.run()



"""
Progress bar segmentation
=========================

Use napari's tqdm wrapper to display the progress of long-running operations
in the viewer.

.. tags:: gui
"""
import numpy as np
from qtpy.QtWidgets import QPushButton, QVBoxLayout, QWidget
from skimage.filters import (
    threshold_isodata,
    threshold_li,
    threshold_otsu,
    threshold_triangle,
    threshold_yen,
)
from skimage.measure import label

import napari
from napari.utils import progress

# we will try each of these thresholds on our image
all_thresholds = [
    threshold_isodata,
    threshold_li,
    threshold_otsu,
    threshold_triangle,
    threshold_yen,
]

viewer = napari.Viewer()

# load cells data and take just nuclei
membrane, cell_nuclei = viewer.open_sample('napari', 'cells3d')
cell_nuclei = cell_nuclei.data


def try_thresholds():
    """Tries each threshold, and adds result to viewer."""
    if 'Binarised' in viewer.layers:
        del viewer.layers['Binarised']

    thresholded_nuclei = []

    # we wrap our iterable with `progress`
    # this will automatically add a progress bar to our activity dock
    for threshold_func in progress(all_thresholds):
        current_threshold = threshold_func(cell_nuclei)
        binarised_im = cell_nuclei > current_threshold
        thresholded_nuclei.append(binarised_im)

        # uncomment if processing is too fast
        # from time import sleep
        # sleep(0.5)

    # working with a wrapped iterable, the progress bar will be closed
    # as soon as the iteration is complete

    binarised_nuclei = np.stack(thresholded_nuclei)
    viewer.add_labels(
        binarised_nuclei,
        color={1: 'lightgreen'},
        opacity=0.7,
        name='Binarised',
        blending='translucent',
    )


# In the previous example, we were able to see the progress bar, but were not
# able to control it. By using `progress` within a context manager, we can
# manipulate the `progress` object and still get the benefit of automatic
# clean up
def segment_binarised_ims():
    """Segments each of the binarised ims.

    Uses `progress` within a context manager allowing us to manipulate
    the progress bar within the loop
    """
    if 'Binarised' not in viewer.layers:
        raise TypeError('Cannot segment before thresholding')
    if 'Segmented' in viewer.layers:
        del viewer.layers['Segmented']
    binarised_data = viewer.layers['Binarised'].data
    segmented_nuclei = []

    # using the `with` keyword we can use `progress` inside a context manager
    # `progress` inherits from tqdm and therefore provides the same API
    # e.g. we can provide the miniters argument if we want to see the
    # progress bar update with each iteration
    with progress(binarised_data, miniters=0) as pbar:
        for i, binarised_cells in enumerate(pbar):
            # this allows us to manipulate the pbar object within the loop
            # e.g. setting the description.
            pbar.set_description(all_thresholds[i].__name__.split('_')[1])
            labelled_im = label(binarised_cells)
            segmented_nuclei.append(labelled_im)

            # uncomment if processing is too fast
            # from time import sleep
            # sleep(0.5)

    # progress bar is still automatically closed

    segmented_nuclei = np.stack(segmented_nuclei)
    viewer.add_labels(
        segmented_nuclei,
        name='Segmented',
        blending='translucent',
    )
    viewer.layers['Binarised'].visible = False


# we can also manually control `progress` objects using their
# `update` method (inherited from tqdm)
def process_ims():
    """
    First performs thresholding, then segmentation on our image.

    Manually updates a `progress` object.
    """
    if 'Binarised' in viewer.layers:
        del viewer.layers['Binarised']
    if 'Segmented' in viewer.layers:
        del viewer.layers['Segmented']

    # we instantiate a manually controlled `progress` object
    # by just passing a total with no iterable
    with progress(total=2) as pbar:
        pbar.set_description('Thresholding')
        try_thresholds()
        # once one processing step is complete, we increment
        # the value of our progress bar
        pbar.update(1)

        pbar.set_description('Segmenting')
        segment_binarised_ims()
        pbar.update(1)

        # uncomment this line to see the 100% progress bar
        # from time import sleep
        # sleep(0.5)

button_layout = QVBoxLayout()
process_btn = QPushButton('Full Process')
process_btn.clicked.connect(process_ims)
button_layout.addWidget(process_btn)

thresh_btn = QPushButton('1.Threshold')
thresh_btn.clicked.connect(try_thresholds)
button_layout.addWidget(thresh_btn)

segment_btn = QPushButton('2.Segment')
segment_btn.clicked.connect(segment_binarised_ims)
button_layout.addWidget(segment_btn)

action_widget = QWidget()
action_widget.setLayout(button_layout)
action_widget.setObjectName('Segmentation')
viewer.window.add_dock_widget(action_widget)

# showing the activity dock so we can see the progress bars
viewer.window._status_bar._toggle_activity_dock(True)

if __name__ == '__main__':
    napari.run()



import os
from enum import auto
from pathlib import Path

from napari.utils.compat import StrEnum


class Environment(StrEnum):
    pip = auto()
    uv = auto()
    conda = auto()
    pixi = auto()


def get_venv_path() -> Path | None:
    """Get the path to the current virtual environment, if any.

    Returns
    -------
    Path | None
        The path to the current virtual environment, or None if not in a venv.
    """
    venv = os.environ.get('VIRTUAL_ENV')
    return Path(venv) if venv else None


def get_conda_path() -> Path | None:
    """Get the path to the current conda environment, if any.

    Returns
    -------
    Path | None
        The path to the current conda environment, or None if not in conda.
    """
    conda = os.environ.get('CONDA_PREFIX')
    return Path(conda) if conda else None


def check_if_uv_env() -> bool:
    """Check if we are in a uv environment.

    Returns
    -------
    bool
        True if we are in a uv virtual environment, False otherwise.
    """
    venv_path = get_venv_path()
    if venv_path is None:
        raise ValueError('Not in a virtual environment')
    if not (venv_path / 'pyvenv.cfg').exists():
        return False
    return 'uv =' in (venv_path / 'pyvenv.cfg').read_text()


def check_if_pixi_env() -> bool:
    """Check if we are in a pixi virtual environment.

    Returns
    -------
    bool
        True if we are in a pixi virtual environment, False otherwise.
    """
    env_path = get_conda_path()
    if env_path is None:
        raise ValueError('Not in a conda environment')
    if not (env_path / 'conda-meta').exists():
        raise ValueError('Not in a conda environment')
    return (env_path / 'conda-meta' / 'pixi_env_prefix').exists()


def detect_environment() -> Environment:
    """Detect the current isolated environment."""
    if get_venv_path() is not None:
        if check_if_uv_env():
            return Environment.uv
        return Environment.pip
    if get_conda_path() is not None:
        if check_if_pixi_env():
            return Environment.pixi
        return Environment.conda
    return Environment.pip



"""
Cursor ray
==========

Depict a ray through a layer in 3D to demonstrate interactive 3D functionality

.. tags:: interactivity
"""
import numpy as np

import napari

sidelength_data = 64
n_points = 10

# data to depict an empty volume, its bounding box and points along a ray
# through the volume
volume = np.zeros(shape=(sidelength_data, sidelength_data, sidelength_data))
bounding_box = np.array(
    [
        [0, 0, 0],
        [1, 0, 0],
        [0, 1, 0],
        [1, 1, 0],
        [0, 0, 1],
        [1, 0, 1],
        [0, 1, 1],
        [1, 1, 1],
    ]
) * sidelength_data
points = np.zeros(shape=(n_points, 3))

# point sizes
point_sizes = np.linspace(0.5, 2, n_points, endpoint=True)

# point colors
green = [0, 1, 0, 1]
magenta = [1, 0, 1, 1]
point_colors = np.linspace(green, magenta, n_points, endpoint=True)

# create viewer and add layers for each piece of data
viewer = napari.Viewer(ndisplay=3)
bounding_box_layer = viewer.add_points(
    bounding_box, face_color='cornflowerblue', name='bounding box'
)
ray_layer = viewer.add_points(
    points, face_color=point_colors, size=point_sizes, name='cursor ray'
)
volume_layer = viewer.add_image(volume, blending='additive')


# callback function, called on mouse click when volume layer is active
@volume_layer.mouse_drag_callbacks.append
def on_click(layer, event):
    near_point, far_point = layer.get_ray_intersections(
        event.position,
        event.view_direction,
        event.dims_displayed
    )
    if (near_point is not None) and (far_point is not None):
        ray_points = np.linspace(near_point, far_point, n_points, endpoint=True)
        if ray_points.shape[1] != 0:
            ray_layer.data = ray_points


if __name__ == '__main__':
    napari.run()



from typing import TYPE_CHECKING

from napari._qt.layer_controls.qt_layer_controls_base import QtLayerControls
from napari._qt.layer_controls.widgets import (
    QtFaceColorControl,
    QtTextVisibilityControl,
)
from napari._qt.layer_controls.widgets._shapes import (
    QtEdgeColorControl,
    QtEdgeWidthSliderControl,
)
from napari._qt.widgets.qt_mode_buttons import QtModePushButton
from napari.layers.shapes._shapes_constants import Mode
from napari.utils.action_manager import action_manager
from napari.utils.interactions import Shortcut
from napari.utils.translations import trans

if TYPE_CHECKING:
    import napari.layers


class QtShapesControls(QtLayerControls):
    """Qt view and controls for the napari Shapes layer.

    Parameters
    ----------
    layer : napari.layers.Shapes
        An instance of a napari Shapes layer.

    Attributes
    ----------
    _edge_color_control : napari._qt.layer_controls.widgets._shapes.QtEdgeColorControl
        Widget that wraps a ColorSwatchEdit controlling current edge color of the layer.
    _edge_width_slider_control : napari._qt.layer_controls.widgets._shapes.QtEdgeWidthSliderControl
        Widget that wraps a slider controlling line edge width of layer.
    _face_color_control : napari._qt.layer_controls.widgets.QtFaceColorControl
        Widget that wraps a ColorSwatchEdit controlling current face color of the layer.
    _text_visibility_control : napari._qt.layer_controls.widgets.QtTextVisibilityControl
        WIdget that wraps a checkbox controlling if text on the layer is visible or not.
    delete_button : qtpy.QtWidgets.QtModePushButton
        Button to delete selected shapes
    direct_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to select individual vertices in shapes.
    ellipse_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to add ellipses to shapes layer.
    move_back_button : qtpy.QtWidgets.QtModePushButton
        Button to move selected shape(s) to the back.
    move_front_button : qtpy.QtWidgets.QtModePushButton
        Button to move shape(s) to the front.
    line_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to add lines to shapes layer.
    path_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to add paths to shapes layer.
    polygon_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to add polygons to shapes layer.
    polygon_lasso_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to add polygons to shapes layer with a lasso tool.
    polyline_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to add polylines to shapes layer.
    rectangle_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to add rectangles to shapes layer.
    select_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to select shapes.
    vertex_insert_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to insert vertex into shape.
    vertex_remove_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to remove vertex from shapes.

    Raises
    ------
    ValueError
        Raise error if shapes mode is not recognized.
    """

    layer: 'napari.layers.Shapes'
    MODE = Mode
    PAN_ZOOM_ACTION_NAME = 'activate_shapes_pan_zoom_mode'
    TRANSFORM_ACTION_NAME = 'activate_shapes_transform_mode'

    def __init__(self, layer) -> None:
        super().__init__(layer)

        # Setup buttons
        self.select_button = self._radio_button(
            layer, 'select', Mode.SELECT, True, 'activate_select_mode'
        )
        self.direct_button = self._radio_button(
            layer, 'direct', Mode.DIRECT, True, 'activate_direct_mode'
        )
        self.rectangle_button = self._radio_button(
            layer,
            'rectangle',
            Mode.ADD_RECTANGLE,
            True,
            'activate_add_rectangle_mode',
        )
        self.ellipse_button = self._radio_button(
            layer,
            'ellipse',
            Mode.ADD_ELLIPSE,
            True,
            'activate_add_ellipse_mode',
        )
        self.line_button = self._radio_button(
            layer, 'line', Mode.ADD_LINE, True, 'activate_add_line_mode'
        )
        self.polyline_button = self._radio_button(
            layer,
            'polyline',
            Mode.ADD_POLYLINE,
            True,
            'activate_add_polyline_mode',
        )
        self.path_button = self._radio_button(
            layer, 'path', Mode.ADD_PATH, True, 'activate_add_path_mode'
        )
        self.polygon_button = self._radio_button(
            layer,
            'polygon',
            Mode.ADD_POLYGON,
            True,
            'activate_add_polygon_mode',
        )
        self.polygon_lasso_button = self._radio_button(
            layer,
            'polygon_lasso',
            Mode.ADD_POLYGON_LASSO,
            True,
            'activate_add_polygon_lasso_mode',
        )
        self.vertex_insert_button = self._radio_button(
            layer,
            'vertex_insert',
            Mode.VERTEX_INSERT,
            True,
            'activate_vertex_insert_mode',
        )
        self.vertex_remove_button = self._radio_button(
            layer,
            'vertex_remove',
            Mode.VERTEX_REMOVE,
            True,
            'activate_vertex_remove_mode',
        )

        self.move_front_button = QtModePushButton(
            layer,
            'move_front',
            tooltip=trans._('Move to front'),
        )
        action_manager.bind_button(
            'napari:move_shapes_selection_to_front', self.move_front_button
        )
        self.move_back_button = QtModePushButton(
            layer,
            'move_back',
            tooltip=trans._('Move to back'),
        )
        action_manager.bind_button(
            'napari:move_shapes_selection_to_back', self.move_back_button
        )
        self.delete_button = QtModePushButton(
            layer,
            'delete_shape',
            slot=self.layer.remove_selected,
            tooltip=trans._(
                'Delete selected shapes ({shortcut})',
                shortcut=Shortcut('Backspace').platform,
            ),
        )
        self._EDIT_BUTTONS += (
            self.delete_button,
            self.move_back_button,
            self.move_front_button,
        )
        self._on_editable_or_visible_change()

        self.button_grid.addWidget(self.move_back_button, 0, 0)
        self.button_grid.addWidget(self.vertex_remove_button, 0, 1)
        self.button_grid.addWidget(self.vertex_insert_button, 0, 2)
        self.button_grid.addWidget(self.delete_button, 0, 3)
        self.button_grid.addWidget(self.direct_button, 0, 4)
        self.button_grid.addWidget(self.select_button, 0, 5)
        self.button_grid.addWidget(self.move_front_button, 1, 0)
        self.button_grid.addWidget(self.ellipse_button, 1, 1)
        self.button_grid.addWidget(self.rectangle_button, 1, 2)
        self.button_grid.addWidget(self.polygon_button, 1, 3)
        self.button_grid.addWidget(self.polygon_lasso_button, 1, 4)
        self.button_grid.addWidget(self.line_button, 1, 5)
        self.button_grid.addWidget(self.polyline_button, 1, 6)
        self.button_grid.addWidget(self.path_button, 1, 7)
        self.button_grid.setContentsMargins(5, 0, 0, 5)
        self.button_grid.setColumnStretch(0, 1)
        self.button_grid.setSpacing(4)

        # Setup widgets controls
        self._edge_width_slider_control = QtEdgeWidthSliderControl(self, layer)
        self._add_widget_controls(self._edge_width_slider_control)
        self._edge_color_control = QtEdgeColorControl(
            self,
            layer,
            tooltip=trans._(
                'Click to set the edge color of currently selected shapes and any added afterwards'
            ),
        )
        self._add_widget_controls(self._edge_color_control)
        self._face_color_control = QtFaceColorControl(
            self,
            layer,
            tooltip=trans._(
                'Click to set the face color of currently selected shapes and any added afterwards.'
            ),
        )
        self._add_widget_controls(self._face_color_control)
        self._text_visibility_control = QtTextVisibilityControl(self, layer)
        self._add_widget_controls(self._text_visibility_control)

    def _on_mode_change(self, event):
        """Update ticks in checkbox widgets when shapes layer mode changed.

        Available modes for shapes layer are:
        * SELECT
        * DIRECT
        * PAN_ZOOM
        * ADD_RECTANGLE
        * ADD_ELLIPSE
        * ADD_LINE
        * ADD_POLYLINE
        * ADD_PATH
        * ADD_POLYGON
        * ADD_POLYGON_LASSO
        * VERTEX_INSERT
        * VERTEX_REMOVE
        * TRANSFORM

        Parameters
        ----------
        event : napari.utils.event.Event
            The napari event that triggered this method.

        Raises
        ------
        ValueError
            Raise error if event.mode is not one of the available modes.
        """
        super()._on_mode_change(event)

    def _on_ndisplay_changed(self):
        self.layer.editable = self.ndisplay == 2
        super()._on_ndisplay_changed()



from qtpy.QtWidgets import (
    QCheckBox,
    QWidget,
)

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import checked_to_bool
from napari.layers import Tracks
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtIdCheckBoxControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the id should be
    displayed attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Tracks
        An instance of a napari Tracks layer.

    Attributes
    ----------
    id_checkbox : qtpy.QtWidgets.QCheckBox
        Checkbox controlling if id of the layer should be shown.
    id_checkbox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for showing the id chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Tracks) -> None:
        super().__init__(parent, layer)
        # Setup layer
        # NOTE(arl): there are no events fired for changing checkbox (layer `display_id` attribute)

        # Setup widgets
        self.id_checkbox = QCheckBox()
        connect_setattr(
            self.id_checkbox.stateChanged,
            layer,
            'display_id',
            convert_fun=checked_to_bool,
        )

        self.id_checkbox_label = QtWrappedLabel(trans._('show ID:'))

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.id_checkbox_label, self.id_checkbox)]



from typing import Any

import gdown
from pooch.downloaders import (
    DOIDownloader,
    FTPDownloader,
    HTTPDownloader,
    SFTPDownloader,
    choose_downloader,
)

DOI_TO_FILE_GOOGLE_ID = {
    "doi:10.5281/zenodo.17668709/ddic.zip": "1-AAkBUXykxXG3Ve20jPk43co-m3f9ZWB",
    "doi:10.5281/zenodo.17668709/grey.zip": "1UZEoOtHMMeJMXVoWr6IZraZnvQe8ExNJ",
    'doi:10.5281/zenodo.13380203/PocilloporaDamicornisSkin.obj': "1yuPHWlLzowlfWzVMUg-mvAEe_Tmvpzy4",
    'doi:10.5281/zenodo.13380203/PocilloporaDamicornisSkin_Texture_0.jpg': "17tG44rMPWjAIoO_AlH9BaQkPY7GxxEN9",
    'doi:10.5281/zenodo.13380203/PocilloporaDamicornisSkin_GeneratedMat2.png': "1l_hGxDg6JARAyFMgWXuoIZs49qKXBv01",

}

GODGLE_PREFIX="https://drive.google.com/open?export=download&id="


class BackupDownloader(HTTPDownloader):
    def __call__(self, url: str, output_file: str, pooch:Any, check_only:bool=False) -> bool | None:
        resource_id = DOI_TO_FILE_GOOGLE_ID[url]
        if check_only:
            download_url = GODGLE_PREFIX + resource_id
            return super().__call__(download_url, output_file, pooch, check_only=check_only)
        gdown.download(id=resource_id, output=output_file, quiet=False)
        return None



def napari_choose_downloader(url: str, progressbar: bool=False) -> HTTPDownloader | SFTPDownloader | FTPDownloader | DOIDownloader:
    if url in DOI_TO_FILE_GOOGLE_ID:
        return BackupDownloader(progressbar=progressbar)
    return choose_downloader(url)



"""
Layer Text Scaling
==================

Display points layer with text features.
By using the `scaling` property of the text layer, the text is scaled with the layers as canvas zoom changes.
Scaling can also be thought of as fixing the height of the text in world coordinates,
whereas no scaling fixes the height of the text in screen / canvas pixels.
This world scaling is useful when there are many points and the text would otherwise overlap when zoomed out, or become illegible when zoomed in.

.. tags:: visualization-advanced
"""

import numpy as np

import napari

viewer = napari.Viewer()
viewer.open_sample('napari', 'astronaut')

# add points with text features
points = np.array([[100, 100], [120, 120], [140, 140], [160, 160]])
text_scaled = {
    'string': ['Point 1', 'Point 2', 'Point 3', 'Point 4'],
    'size': 8,
    'color': 'magenta',
    'anchor': 'center',
    'scaling': True
}
points_layer = viewer.add_points(
    points,
    size=20,
    text=text_scaled,
    name='scaled_text (magenta)'
)
# points_layer.text.scaling = True  # alternative way to set scaling

# add a second points layer slightly offset and show without scaling
text_unscaled = {
    'string': ['Point 1', 'Point 2', 'Point 3', 'Point 4'],
    'size': 8,
    'color': 'green',
    'anchor': 'center',
}
points_layer2 = viewer.add_points(
    points + [40, 0],
    size=20,
    text=text_unscaled,
    name='unscaled_text (green)'
)

shape_text = {
    'string': ['Triangle'],
    'size': 12,
    'scaling': True,
}
shapes_layer = viewer.add_shapes(
    data=[[100, 300], [200, 300], [100, 400]],
    shape_type=['polygon'],
    text=shape_text,
    name='scaled_text_shapes'
)

for layer in viewer.layers:
    layer.scale = (3, 3)

viewer.reset_view()

if __name__ == '__main__':
    napari.run()



"""
Export Figure
=============

Display a variety of layer types in the napari viewer and export the figure with `viewer.export_figure()`.
The exported figure is then added back as an image layer.

Exported figures include the extent of all data in 2D or 3D view.
To capture the extent of the canvas, instead of the layers, see `viewer.screenshot()`: :ref:`sphx_glr_gallery_to_screenshot.py` and :ref:`sphx_glr_gallery_screenshot_and_export_figure.py`.

.. tags:: visualization-advanced
"""

import numpy as np
from skimage import data

import napari

# create the viewer and window
viewer = napari.Viewer()

# add the image
img_layer = viewer.add_image(data.camera(), name='photographer')
img_layer.colormap = 'gray'

# create a list of polygons
polygons = [
    np.array([[11, 13], [111, 113], [22, 246]]),
    np.array(
        [
            [505, 60],
            [402, 71],
            [383, 42],
            [251, 95],
            [212, 59],
            [131, 137],
            [126, 187],
            [191, 204],
            [171, 248],
            [211, 260],
            [273, 243],
            [264, 225],
            [430, 173],
            [512, 160],
        ]
    ),
    np.array(
        [
            [310, 382],
            [229, 381],
            [209, 401],
            [221, 411],
            [258, 411],
            [300, 412],
            [306, 435],
            [268, 434],
            [265, 454],
            [298, 461],
            [307, 461],
            [307, 507],
            [349, 510],
            [352, 369],
            [330, 366],
            [330, 366],
        ]
    ),
]

# add polygons
layer = viewer.add_shapes(
    polygons,
    shape_type='polygon',
    edge_width=1,
    edge_color='coral',
    face_color='royalblue',
    name='shapes',
)

# add an ellipse to the layer
ellipse = np.array([[59, 222], [110, 289], [170, 243], [119, 176]])
layer.add(
    ellipse,
    shape_type='ellipse',
    edge_width=5,
    edge_color='coral',
    face_color='purple',
)

labels = layer.to_labels([512, 512])
labels_layer = viewer.add_labels(labels, name='labels')

points = np.array([[100, 100], [200, 200], [333, 111]])
size = np.array([10, 20, 20])
viewer.add_points(points, size=size)

# Add scale bar of a defined length to the exported figure
viewer.scale_bar.visible = True
viewer.scale_bar.length = 250

# Export figure and change theme before and after exporting to show that the background canvas margins
# are not in the exported figure.
viewer.theme = "light"
# Optionally for saving the exported figure: viewer.export_figure(path="export_figure.png")
export_figure = viewer.export_figure()
scaled_export_figure = viewer.export_figure(scale_factor=5)
viewer.theme = "dark"

viewer.add_image(export_figure, rgb=True, name='exported_figure')
viewer.add_image(scaled_export_figure, rgb=True, name='scaled_exported_figure')
viewer.reset_view()

if __name__ == '__main__':
    napari.run()



"""
nD multiscale image
===================

Displays an nD multiscale image

.. tags:: visualization-advanced
"""

import numpy as np
from skimage.transform import pyramid_gaussian

import napari

# create multiscale from random data
base = np.random.random((1536, 1536))
base = np.array([base * (8 - i) / 8 for i in range(8)])
print('base shape', base.shape)
multiscale = list(
    pyramid_gaussian(base, downscale=2, max_layer=2, channel_axis=-1)
)
print('multiscale level shapes: ', [p.shape for p in multiscale])

# add image multiscale
viewer = napari.Viewer()
layer = viewer.add_image(multiscale, contrast_limits=[0, 1], multiscale=True)

if __name__ == '__main__':
    napari.run()



"""
Viewer FPS label
================

Display a 3D volume and the fps label.

.. tags:: experimental
"""
import numpy as np

import napari


def update_fps(fps):
    """Update fps."""
    viewer.text_overlay.text = f'{fps:1.1f} FPS'


viewer = napari.Viewer()
viewer.add_image(np.random.random((5, 5, 5)), colormap='red', opacity=0.8)
viewer.text_overlay.visible = True
# note: this is using a private attribute, so it might break
# without warning in future versions!
viewer.window._qt_viewer.canvas._scene_canvas.measure_fps(callback=update_fps)

# call update_fps function once to have it show before
# mouse interaction with the canvas
update_fps(60)
if __name__ == '__main__':
    napari.run()



"""
Dynamic projections dask
========================

Using dask array operations, one can dynamically take arbitrary slices
and computations of a source dask array and display the results in napari.
When the computation takes one or more parameters, one can tie a UI to
them using magicgui.

.. tags:: visualization-advanced
"""

import dask.array as da
import numpy as np
from dask.array.lib.stride_tricks import sliding_window_view
from skimage import data

import napari

##############################################################################
# Part 1: using code to view a specific value.

blobs = data.binary_blobs(length=64, n_dim=3)
blobs_dask = da.from_array(blobs, chunks=(1, 64, 64))

# original shape [60, 1, 1, 5, 64, 64],
# use squeeze to remove singleton axes
blobs_dask_windows = np.squeeze(
    sliding_window_view(blobs_dask, window_shape=(5, 64, 64)),
    axis=(1, 2),
)
blobs_sum = np.sum(blobs_dask_windows, axis=1)
viewer = napari.Viewer()
layer = viewer.add_image(blobs_sum)

if __name__ == '__main__':
    napari.run()

##############################################################################
# Part 2: using magicgui to vary the slice thickness.

from magicgui import magicgui  # noqa: E402


def sliding_window_mean(
    arr: napari.types.ImageData, size: int = 1
) -> napari.types.LayerDataTuple:
    window_shape = (size,) + (arr.shape[1:])
    arr_windows = sliding_window_view(arr, window_shape=window_shape)
    # as before, use squeeze to remove singleton axes
    arr_windows_1d = np.squeeze(
        arr_windows, axis=tuple(range(1, arr.ndim))
    )
    arr_summed = np.sum(arr_windows_1d, axis=1) / size
    return (
        arr_summed,
        {
            'translate': (size // 2,) + (0,) * (arr.ndim - 1),
            'name': 'mean-window',
            'colormap': 'magenta',
            'blending': 'additive',
        },
        'image',
    )


viewer = napari.Viewer()
layer = viewer.add_image(blobs_dask, colormap='green')
viewer.window.add_dock_widget(magicgui(sliding_window_mean, auto_call=True))
viewer.dims.current_step = (32, 0, 0)

if __name__ == '__main__':
    napari.run()



from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QWidget,
)
from superqt import QLargeIntSpinBox

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import attr_to_settr
from napari.layers import Labels
from napari.layers.labels._labels_utils import get_dtype
from napari.utils._dtype import get_dtype_limits
from napari.utils.translations import trans


class QtContourSpinBoxControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer contour
    thickness attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Labels
        An instance of a napari Labels layer.

    Attributes
    ----------
    contour_spinbox : superqt.QLargeSpinBox
        Spinbox to control the layer contour thickness.
    contour_spinbox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the layer contour thickness chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Labels) -> None:
        super().__init__(parent, layer)
        # Setup widgets
        self.contour_spinbox = QLargeIntSpinBox()
        dtype_lims = get_dtype_limits(get_dtype(layer))
        self.contour_spinbox.setRange(0, dtype_lims[1])
        self.contour_spinbox.setToolTip(
            trans._('Set width of displayed label contours')
        )
        self.contour_spinbox.setValue(self._layer.contour)
        self.contour_spinbox.valueChanged.connect(self.change_contour)
        self.contour_spinbox.setKeyboardTracking(False)
        self.contour_spinbox.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._callbacks.append(
            attr_to_settr(
                self._layer,
                'contour',
                self.contour_spinbox,
                'setValue',
            )
        )

        self.contour_spinbox_label = QtWrappedLabel(trans._('contour:'))

    def change_contour(self, value: int) -> None:
        """Change contour thickness.
        Parameters
        ----------
        value : int
            Thickness of contour.
        """
        self._layer.contour = value
        self.contour_spinbox.clearFocus()
        self.parent().setFocus()

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.contour_spinbox_label, self.contour_spinbox)]



"""
Live tiffs
==========

Loads and Displays tiffs as they get generated in the specific directory.
Trying to simulate the live display of data as it gets acquired by microscope.
This script should be run together with live_tiffs_generator.py

.. tags:: experimental
"""

import os
import sys
import time

import dask.array as da
from dask import delayed
from skimage.io.collection import alphanumeric_key
from tifffile import imread

import napari
from napari.qt import thread_worker

viewer = napari.Viewer(ndisplay=3)
# pass a directory to monitor or it will monitor current directory.
path = sys.argv[1] if len(sys.argv) > 1 else '.'
path = os.path.abspath(path)
end_of_experiment = 'final.log'


def append(delayed_image):
    """Appends the image to viewer.

    Parameters
    ----------
    delayed_image : dask.delayed function object
    """
    if delayed_image is None:
        return

    if viewer.layers:
        # layer is present, append to its data
        layer = viewer.layers[0]
        image_shape = layer.data.shape[1:]
        image_dtype = layer.data.dtype
        image = da.from_delayed(
            delayed_image, shape=image_shape, dtype=image_dtype,
        ).reshape((1, *image_shape))
        layer.data = da.concatenate((layer.data, image), axis=0)
    else:
        # first run, no layer added yet
        image = delayed_image.compute()
        image = da.from_delayed(
            delayed_image, shape=image.shape, dtype=image.dtype,
        ).reshape((1, *image.shape))
        layer = viewer.add_image(image, rendering='attenuated_mip')

    # we want to show the last file added in the viewer to do so we want to
    # put the slider at the very end. But, sometimes when user is scrolling
    # through the previous slide then it is annoying to jump to last
    # stack as it gets added. To avoid that jump we 1st check where
    # the scroll is and if its not at the last slide then don't move the slider.
    if viewer.dims.point[0] >= layer.data.shape[0] - 2:
        viewer.dims.set_point(0, layer.data.shape[0] - 1)


@thread_worker(connect={'yielded': append})
def watch_path(path):
    """Watches the path for new files and yields it once file is ready.

    Notes
    -----
    Currently, there is no proper way to know if the file has written
    entirely. So the workaround is we assume that files are generating
    serially (in most microscopes it common), and files are name in
    alphanumeric sequence We start loading the total number of minus the
    last file (`total__files - last`). In other words, once we see the new
    file in the directory, it means the file before it has completed so load
    that file. For this example, we also assume that the microscope is
    generating a `final.log` file at the end of the acquisition, this file
    is an indicator to stop monitoring the directory.

    Parameters
    ----------
    path : str
    directory to monitor and load tiffs as they start appearing.
    """
    current_files = set()
    processed_files = set()
    end_of_acquisition = False
    while not end_of_acquisition:
        files_to_process = set()
        # Get the all files in the directory at this time
        current_files = set(os.listdir(path))

        # Check if the end of acquisition has reached
        # if yes then remove it from the files_to_process set
        # and send it to display
        if end_of_experiment in current_files:
            files_to_process = current_files - processed_files
            files_to_process.remove(end_of_experiment)
            end_of_acquisition = True

        elif len(current_files):
            # get the last file from the current files based on the file names
            last_file = sorted(current_files, key=alphanumeric_key)[-1]
            current_files.remove(last_file)
            files_to_process = current_files - processed_files

        # yield every file to process as a dask.delayed function object.
        for p in sorted(files_to_process, key=alphanumeric_key):
            yield delayed(imread)(os.path.join(path, p))
        else:
            yield

        # add the files which we have yield to the processed list.
        processed_files.update(files_to_process)
        time.sleep(0.1)


worker = watch_path(path)
if __name__ == '__main__':
    napari.run()



"""
Cursor position
===============

Add small data to examine cursor positions

.. tags:: interactivity
"""

import numpy as np

import napari

viewer = napari.Viewer()
image = np.array([[1, 0, 0, 1],
                  [0, 0, 1, 1],
                  [1, 0, 3, 0],
                  [0, 2, 0, 0]], dtype=int)

viewer.add_labels(image)

points = np.array([[0, 0], [2, 0], [1, 3]])
viewer.add_points(points, size=0.25)

rect = np.array([[0, 0], [3, 1]])
viewer.add_shapes(rect, shape_type='rectangle', edge_width=0.1)

vect = np.array([[[3, 2], [-1, 1]]])
viewer.add_vectors(vect, edge_width=0.1)

if __name__ == '__main__':
    napari.run()



import time
import warnings

import napari
from napari._qt.widgets.qt_viewer_buttons import QtViewerPushButton
from napari.qt import thread_worker


@thread_worker(start_thread=True)
def make_warning(*_):
    time.sleep(0.05)
    warnings.warn('Warning in another thread')


@thread_worker(start_thread=True)
def make_error(*_):
    time.sleep(0.05)
    raise ValueError('Error in another thread')


viewer = napari.Viewer()
layer_buttons = viewer.window.qt_viewer.layerButtons
err_btn = QtViewerPushButton(None, 'warning', 'new Error', make_error)
warn_btn = QtViewerPushButton(None, 'warning', 'new Warn', make_warning)
layer_buttons.layout().insertWidget(3, warn_btn)
layer_buttons.layout().insertWidget(3, err_btn)

if __name__ == '__main__':
    napari.run()



"""Visualize the triangulation algorithms used by the Shapes layer.

This example uses napari layers to draw each of the components of a face and
edge triangulation in a Shapes layer.

Shapes layers don't "just" draw polygons, ellipses, and so on: each shape,
as well as its borders, is broken down into a collection of triangles (this
is called a *triangulation*), which are then sent to OpenGL for drawing:
drawing triangles is one of the "visualization primitives" in OpenGL and most
2D and 3D drawing frameworks.

It turns out that converting arbitrary shapes into a collection of triangles
can be quite tricky: very shallow angles cause errors in the algorithms, and
can also make certain desirable properties (such as edges not overlapping with
each other when a polygon makes a sharp turn) actually impossible to achieve
with fast (single-pass) algorithms.

This example draws the Shapes layer normally, but also overlays all the
elements of the triangulation: the triangles themselves, and the normal vectors
on each polygon vertex, from which the triangulation is computed.
"""

from dataclasses import dataclass, fields
from functools import partial

try:
    import numba
except ImportError:
    # Create a dummy numba.njit to allow running this script without numba
    import toolz as tz
    class numba:
        @tz.curry
        def njit(func, cache=True, inline=None):
            return func

import numpy as np

import napari
from napari.layers import Shapes
from napari.layers.shapes._accelerated_triangulate_python import (
    generate_2D_edge_meshes_py,
)


def generate_regular_polygon(n, radius=1):
    """Generate vertices of a regular n-sided polygon centered at the origin.

    Parameters
    ----------
    n : int
        The number of sides (vertices).
    radius : float, optional
        The radius of the circumscribed circle.

    Returns
    -------
    np.ndarray
        An array of shape (n, 2) containing the vertex coordinates.
    """
    angles = np.linspace(0, 2 * np.pi, n, endpoint=False)
    return np.column_stack((radius * np.cos(angles), radius * np.sin(angles)))


def get_reference_edge_triangulation_points(shape: Shapes) -> np.ndarray:
    """Get the non-accelerated points"""
    shapes = shape._data_view.shapes
    path_list = [(x.data, x._closed) for x in shapes]
    mesh_list = [generate_2D_edge_meshes_py(path, closed)
                 for path, closed in path_list]
    mesh = tuple(np.concatenate(el, axis=0) for el in zip(*mesh_list, strict=False))

    return mesh[0] + mesh[1]


def generate_order_vectors(path_, closed) -> np.ndarray:
    """Generate the vectors tangent to the path.

    Parameters
    ----------
    path_ : np.ndarray, shape (N, 2)
        A list of 2D path coordinates.
    closed : bool
        Whether the coordinates represent a closed polygon or an open
        line/path.

    Returns
    -------
    vec : np.ndarray, shape (N, 2, 2)
        A set of vectors, defined by a 2D position and a 2D projection.
    """
    raw_vecs = np.diff(path_, axis=0)
    norm = np.linalg.norm(raw_vecs, axis=1, keepdims=True)
    directions = raw_vecs / norm
    vec = np.empty((path_.shape[0], 2, 2))
    vec[:, 0] = path_
    vec[:-1, 1] = directions
    if closed:
        # point from the last vertex towards the first vertex
        vec[-1, 1] = (
                (path_[0] - path_[-1]) / np.linalg.norm(path_[-1] - path_[0])
                )
    else:
        # point back towards the penultimate vertex
        vec[-1, 1] = -vec[-2, 1]
    return vec


def generate_miter_helper_vectors(
        direction_vectors_array: np.ndarray
        ) -> np.ndarray:
    """Generate the miter helper vectors.

    The miter helper vectors are a pair of vectors for each point in the path,
    which together help define whether a bevel join will be needed. The first
    vector is half of the normalized direction vector for that vertex, and the
    second is *minus* half of the normalized direction vector for the previous
    vertex. Their angle is the angle of the edges at that vertex.

    Parameters
    ----------
    direction_vectors_array : array of shape (n, 2)
        array of normalized (length 1) direction vectors for each point in the
        path.

    Returns
    -------
    array of shape (n, 2, 2)
        array of miter helper vectors
    """
    half_direction = direction_vectors_array.copy()
    half_direction[:, 1] *= 0.5
    half_prev_direction = half_direction.copy()
    half_prev_direction[:, 1] *= -1
    half_prev_direction[:, 1] = np.roll(half_prev_direction[:, 1], 1, axis=0)
    half_prev_direction[:, 0] += half_direction[:, 1]
    return np.concatenate([half_direction, half_prev_direction], axis=0)


@numba.njit
def generate_orthogonal_vectors(direction_vectors: np.ndarray) -> np.ndarray:
    """Generate the orthogonal vectors to the direction vectors.

    The orthogonal vector starts at the middle of the direction vector and is
    orthogonal to it in the positive orientation. Its length is half of the
    direction vector, to align with the normalized edge width.

    Parameters
    ----------
    direction_vectors : array, shape (n, 2, 2)
        The direction vector start points (``direction_vectors[:, 0, :]``) and
        directions (``direction_vectors[:, 1, :]``).

    Returns
    -------
    orthogonal_vectors : array, shape(n, 2, 2)
        The orthogonal vector start points and directions.
    """
    position = 0
    vector = 1
    y, x = 0, 1
    half_direction = 0.5 * direction_vectors[:, 1, :]
    orthogonal_vectors = direction_vectors.copy()
    orthogonal_vectors[:, position] = (
            direction_vectors[:, position] + half_direction
            )

    orthogonal_vectors[:, vector, y] = -half_direction[:, x]
    orthogonal_vectors[:, vector, x] = half_direction[:, y]
    return orthogonal_vectors


@numba.njit
def generate_miter_vectors(
        mesh: tuple[np.ndarray, np.ndarray, np.ndarray]
        ) -> np.ndarray:
    """For each point on path, generate the vectors pointing to miter points.

    Parameters
    ----------
    mesh : tuple[np.ndarray]
        vertices, offsets, and triangles of the mesh corresponding to the edges
        of a single shape.

    Returns
    -------
    np.ndarray, shape (n, 2, 2)
        Positions and projections of vectors corresponding to the miter points
        offset from the path points.
    """
    vec_points = np.empty((mesh[0].shape[0], 2, 2))
    vec_points[:, 0] = mesh[0]
    vec_points[:, 1] = mesh[1]
    return vec_points


@numba.njit
def generate_edge_triangle_borders(centers, offsets, triangles) -> np.ndarray:
    """Generate 3 vectors that represent the borders of the triangle.

    These are vectors to show the *ordering* of the triangles in the data.

    Parameters
    ----------
    centers, offsets, triangles : np.ndarray of float
        The mesh triangulation of the shape's edge path.

    Returns
    -------
    borders : np.ndarray of float
        Positions and projections corresponding to each triangle edge in
        the triangulation of the path.
    """
    borders = np.empty((len(triangles)*3, 2, 2), dtype=centers.dtype)
    position = 0
    projection = 1
    for i, triangle in enumerate(triangles):
        a, b, c = triangle
        a1 = centers[a] + offsets[a]
        b1 = centers[b] + offsets[b]
        c1 = centers[c] + offsets[c]
        borders[i * 3, position] = a1
        borders[i * 3, projection] = (b1 - a1)
        borders[i * 3 + 1, position] = b1
        borders[i * 3 + 1, projection] = (c1 - b1)
        borders[i * 3 + 2, position] = c1
        borders[i * 3 + 2, projection] = (a1 - c1)
    return borders


@numba.njit
def generate_face_triangle_borders(vertices, triangles) -> np.ndarray:
    """For each triangle in mesh generate 3 vectors that represent the borders of the triangle.
    """
    borders = np.empty((len(triangles)*3, 2, 2), dtype=vertices.dtype)
    for i, triangle in enumerate(triangles):
        a, b, c = triangle
        a1 = vertices[a]
        b1 = vertices[b]
        c1 = vertices[c]
        borders[i * 3, 0] = a1
        borders[i * 3, 1] = (b1 - a1)
        borders[i * 3 + 1, 0] = b1
        borders[i * 3 + 1, 1] = (c1 - b1)
        borders[i * 3 + 2, 0] = c1
        borders[i * 3 + 2, 1] = (a1 - c1)
    return borders


@dataclass
class Helpers:
    """Simple class to hold all auxiliary vector data for a shapes layer."""
    reference_join_points: np.ndarray
    join_points: np.ndarray
    direction_vectors: np.ndarray
    miter_helper_vectors: np.ndarray
    orthogonal_vectors: np.ndarray
    miter_vectors: np.ndarray
    triangles_vectors: np.ndarray
    face_triangles_vectors: np.ndarray


def asdict(dataclass_instance):
    """Shallow copy version of `dataclasses.asdict`."""
    return {f.name: getattr(dataclass_instance, f.name)
            for f in fields(dataclass_instance)}


def get_helper_data_from_shapes(shapes_layer: Shapes) -> Helpers:
    """Function to generate all auxiliary data for a shapes layer."""
    shapes = shapes_layer._data_view.shapes
    mesh_list = [(s._edge_vertices, s._edge_offsets, s._edge_triangles)
                 for s in shapes]
    path_list = [(s.data, s._closed) for s in shapes]
    mesh = tuple(np.concatenate(el, axis=0) for el in zip(*mesh_list, strict=False))
    face_mesh_list = [(s._face_vertices, s._face_triangles)
                      for s in shapes
                      if len(s._face_vertices) > 0]
    order_vectors_list = [generate_order_vectors(path_, closed)
                          for path_, closed in path_list]

    helpers = Helpers(
        reference_join_points=get_reference_edge_triangulation_points(
                shapes_layer
                ),
        join_points=mesh[0] + mesh[1],
        direction_vectors=np.concatenate(order_vectors_list, axis=0),
        miter_helper_vectors=np.concatenate(
                [generate_miter_helper_vectors(o) for o in order_vectors_list],
                axis=0,
                ),
        orthogonal_vectors=np.concatenate(
                [generate_orthogonal_vectors(o) for o in order_vectors_list],
                axis=0,
                ),
        miter_vectors=np.concatenate(
                [generate_miter_vectors(m) for m in mesh_list],
                axis=0,
                ),
        triangles_vectors=np.concatenate(
                [generate_edge_triangle_borders(*m) for m in mesh_list],
                axis=0,
                ),
        face_triangles_vectors=np.concatenate(
                [generate_face_triangle_borders(*m) for m in face_mesh_list],
                axis=0,
                ),
    )

    return helpers


def update_helper_layers(viewer: napari.Viewer, source_layer: Shapes):
    updated_helpers = get_helper_data_from_shapes(source_layer)
    for name, data in asdict(updated_helpers).items():
        viewer.layers[name].data = data


def add_helper_layers(viewer: napari.Viewer, source_layer):
    """Add helper layers to the viewer that track with the source shapes."""
    helpers = get_helper_data_from_shapes(source_layer)
    # sizes and colors are hardcoded based on vibes
    sizes = [0.2, 0.1, 0.1, 0.06, 0.04, 0.05, 0.04, 0.04]
    colors = ['yellow', 'white', 'red', 'blue',
              'green', 'yellow', 'pink', 'magenta']
    for (name, data), size, color in zip(
            asdict(helpers).items(), sizes, colors, strict=False
            ):
        if 'points' in name:
            viewer.add_points(data, name=name, size=size, face_color=color)
        else:  # all other fields are vectors
            viewer.add_vectors(
                    data,
                    name=name,
                    vector_style='arrow', edge_width=size, edge_color=color,
                    )
    source_layer.events.set_data.connect(
        partial(update_helper_layers, viewer=viewer, source_layer=source_layer)
    )

def pentagram():
    radius = 10
    n = 5
    angles = np.linspace(0, 4 * np.pi, n, endpoint=False)
    return np.column_stack((radius * np.cos(angles), radius * np.sin(angles)))



path = np.array([[0,0], [0,1], [1,1], [1,0]]) * 10
sparkle = np.array([[1, 1], [10, 0], [1, -1], [0, -10],
                    [-1, -1], [-10, 0], [-1, 1], [0, 10]])
fork = np.array([[2, 10], [0, -5], [-2, 10], [-2, -10], [2, -10]])

poly_hole = np.array([[0,0], [10, 0], [10, 10], [0, 10], [0, 0], [2, 5], [5, 8], [8, 5], [5, 2], [2, 5]]) *1.5


polygons = [
    # square
    generate_regular_polygon(4, radius=1) * 10,
    # decagon
    generate_regular_polygon(10, radius=1) * 10 + np.array([[25, 0]]),
    # triangle
    generate_regular_polygon(3, radius=1) * 10 + np.array([[0, 25]]),
    # two sharp prongs
    fork + np.array([[25, 25]]),
    # a four-sided star
    sparkle + np.array([[50, 0]]),
    # same star, but winding in the opposite direction
    sparkle[::-1] + np.array([[50, 26]]),
    # problem shape —
    # lighting bolt with sharp angles and overlapping edge widths
    np.array([[10.97627008, 14.30378733],
              [12.05526752, 10.89766366],
              [8.47309599, 12.91788226],
              [8.75174423, 17.83546002],
              [19.27325521, 7.66883038],
              [15.83450076, 10.5778984]],
             ) + np.array([[60, -15]]),
    poly_hole + np.array([[65, 20]]),
    pentagram() + np.array([[70, 75]]),
    ]

paths = [
    # a simple backwards-c shape
    path + np.array([[0, 50]]),
    # an unclosed decagon
    generate_regular_polygon(10, radius=1) * 10 + np.array([[25, 50]]),
    # a three-point straight line (tests collinear points in path)
    np.array([[0, -10], [0, 0], [0, 10]]) + np.array([[50, 50]]),
    poly_hole + np.array([[65, 45]]),
    ]

shapes = polygons + paths
shape_types=['polygon'] * len(polygons) + ['path'] * len(paths)

viewer = napari.Viewer()
shapes_layer = viewer.add_shapes(shapes, shape_type=shape_types, name='shapes')

add_helper_layers(viewer, source_layer=shapes_layer)
viewer.layers.selection = {shapes_layer}
viewer.reset_view()

if __name__ == '__main__':
    napari.run()



import numpy as np
from vispy.scene.visuals import Compound, Line, Rectangle

from napari._vispy.visuals.text import Text


class ScaleBar(Compound):
    """Scale bar visual with text and line components.

    Layout of Scale Bar Elements (from top to bottom):
    - Padding
    - Text
    - Gap between text and line
    - Scale line (with optional ticks)
    - Padding
    """

    def __init__(self) -> None:
        # Layout constants
        self.PADDING = 6  # Space around the entire scale bar
        self.TICK_LENGTH = 11  # Height of tick marks (odd numbers look better)

        # Line geometry: main line + optional tick marks
        self._line_data = np.array(
            [
                [-1, 0],  # Left end of main line
                [1, 0],  # Right end of main line
                [-1, -1],  # Left tick mark (bottom)
                [-1, 1],  # Left tick mark (top)
                [1, -1],  # Right tick mark (bottom)
                [1, 1],  # Right tick mark (top)
            ]
        )

        self._color = (1, 1, 1, 1)
        self._box_color = (0, 0, 0, 1)

        self.box = Rectangle(center=[0.5, 0.5], width=100, height=36)
        self.text = Text(
            text='1px',
            pos=[0.5, 0.5],
            anchor_x='center',
            anchor_y='bottom',
            font_size=10,
        )
        self.line = Line(
            connect='segments', method='gl', width=3, antialias=True
        )
        # order matters (last is drawn on top)
        super().__init__([self.box, self.text, self.line])

    def _calculate_layout(self, length: float) -> dict:
        """Calculate all layout dimensions and positions."""
        # Text dimensions
        text_width, text_height = self.text.get_width_height()
        # add some extra padding between the scale bar and text
        text_height *= 1.5

        # Box dimensions
        box_width = max(
            length + 2 * self.PADDING,  # Space for line + padding
            text_width + 2 * self.PADDING,  # Space for text + padding
        )
        box_height = (
            self.PADDING  # Top padding
            + text_height  # Text height
            + self.TICK_LENGTH  # Line + ticks height
            + self.PADDING  # Bottom padding
        )

        # Element positions (Y coordinates from top of box)
        text_y = self.PADDING
        line_center_y = self.PADDING + text_height + (self.TICK_LENGTH / 2)

        return {
            'box_width': box_width,
            'box_height': box_height,
            'text_y': text_y,
            'line_center_y': line_center_y,
        }

    def set_data(self, *, length, color, ticks, font_size):
        """Update scale bar with new dimensions and styling."""
        # font size need to be set first cause layout calculations depend on it
        self.text.font_size = font_size

        layout = self._calculate_layout(length)

        # Choose line data based on whether ticks are enabled
        line_data = self._line_data if ticks else self._line_data[:2]

        # Position and scale the line
        self.line.set_data(
            pos=line_data * (length / 2, self.TICK_LENGTH / 2)
            + (
                layout['box_width'] / 2,  # Center horizontally
                layout['line_center_y'],  # Position vertically
            ),
            color=color,
        )

        # Set up the background box
        self.box.width = layout['box_width']
        self.box.height = layout['box_height']
        self.box.center = layout['box_width'] / 2, layout['box_height'] / 2

        # Position the text
        self.text.pos = layout['box_width'] / 2, layout['text_y']
        self.text.color = color

        # Return dimensions for the overlay system
        # Extra padding needed for proper canvas positioning (not sure why padding is needed here, ugh)
        return layout['box_width'], layout['box_height'] + self.PADDING



"""
Clipboard
=========

Copy screenshot of the canvas or the whole viewer to clipboard.

.. tags:: gui
"""

from qtpy.QtWidgets import QPushButton, QVBoxLayout, QWidget
from skimage import data

import napari

# create the viewer with an image
viewer = napari.Viewer()
layer = viewer.add_image(data.moon())

class Grabber(QWidget):
    def __init__(self) -> None:
        super().__init__()

        self.copy_canvas_btn = QPushButton('Copy Canvas to Clipboard', self)
        self.copy_canvas_btn.setToolTip('Copy screenshot of the canvas to clipboard.')
        self.copy_viewer_btn = QPushButton('Copy Viewer to Clipboard', self)
        self.copy_viewer_btn.setToolTip('Copy screenshot of the entire viewer to clipboard.')

        layout = QVBoxLayout(self)
        layout.addWidget(self.copy_canvas_btn)
        layout.addWidget(self.copy_viewer_btn)


def create_grabber_widget():
    """Create widget"""
    widget = Grabber()

    # connect buttons
    widget.copy_canvas_btn.clicked.connect(lambda: viewer.window.clipboard(canvas_only=True))
    widget.copy_viewer_btn.clicked.connect(lambda: viewer.window.clipboard(canvas_only=False))
    return widget


widget = create_grabber_widget()
viewer.window.add_dock_widget(widget)

if __name__ == '__main__':
    napari.run()



"""
Tiled rendering 2D
==================

This example shows how to display tiled, chunked data in napari using the
experimental octree support.

If given a large 2D image with octree support enabled, napari will only load
and display the tiles in the center of the current canvas view. (Note: napari
uses its own internal tile size that may or may not be aligned with the
underlying tiled data, but this should have only minor performance
consequences.)

If octree support is *not* enabled, napari will try to load the entire image,
which may not fit in memory and may bring your computer to a halt. Oops! So, we
make sure that we enable octree support by setting the NAPARI_OCTREE
environment variable to 1 if it is not set by the user.

.. tags:: experimental
"""

import os

# important: if this is not set, the entire ~4GB array will be created!
os.environ.setdefault('NAPARI_OCTREE', '1')

import dask.array as da

import napari

ndim = 2
data = da.random.randint(
        0, 256, (65536,) * ndim,
        chunks=(256,) * ndim,
        dtype='uint8'
        )

viewer = napari.Viewer()
viewer.add_image(data, contrast_limits=[0, 255])
# To turn off grid lines
#viewer.layers[0].display.show_grid = False

# set small zoom so we don't try to load the whole image at once
viewer.camera.zoom = 0.75

# run the example — try to pan around!
if __name__ == '__main__':
    napari.run()



"""
Linked layers
=============

Demonstrates the `link_layers` function.

This function takes a list of layers and an optional list of attributes, and
links them such that when one of the linked attributes changes on any of the
linked layers, all of the other layers follow.

.. tags:: experimental
"""
import numpy as np

import napari
from napari.experimental import link_layers

viewer = napari.Viewer()
layer = viewer.add_image(np.random.rand(3, 64, 64), channel_axis=0)

# link contrast_limits and gamma between all layers in viewer
# NOTE: you may also omit the second argument to link ALL valid, common
# attributes for the set of layers provided
link_layers(viewer.layers, ('contrast_limits', 'gamma'))

# unlinking may be done with napari.experimental.unlink_layers

# this may also be done in a context manager:
# with napari.experimental.layers_linked([layers]):
#     ...

if __name__ == '__main__':
    napari.run()



"""
magicgui parameter sweep
========================

Example showing how to accomplish a napari parameter sweep with magicgui.

It demonstrates:
1. overriding the default widget type with a custom class
2. the `auto_call` option, which calls the function whenever a parameter changes

.. tags:: gui
"""
from typing import Annotated

import skimage.data
import skimage.filters

import napari


# Define our gaussian_blur function.
# Note that we can use forward references for the napari type annotations.
# You can read more about them here:
# https://peps.python.org/pep-0484/#forward-references
# In this example, because we have already imported napari anyway, it doesn't
# really matter. But this syntax would let you specify that a parameter is a
# napari object type without actually importing or depending on napari.
# We also use the `Annotated` type to pass an additional dictionary that can be used
# to aid widget generation. The keys of the dictionary are keyword arguments to
# the corresponding magicgui widget type. For more information see
# https://napari.org/magicgui/api/widgets.html.
def gaussian_blur(
    layer: 'napari.layers.Image',
    sigma: Annotated[float, {'widget_type': 'FloatSlider', 'max': 6}] = 1.0,
    mode: Annotated[str, {'choices': ['reflect', 'constant', 'nearest', 'mirror', 'wrap']}]='nearest',
) -> 'napari.types.ImageData | None':
    """Apply a gaussian blur to ``layer``."""
    if layer:
        return skimage.filters.gaussian(layer.data, sigma=sigma, mode=mode)
    return None


# create a viewer and add some images
viewer = napari.Viewer()
viewer.add_image(skimage.data.astronaut().mean(-1), name='astronaut')
viewer.add_image(skimage.data.grass().astype('float'), name='grass')

# Add our magic function to napari
viewer.window.add_function_widget(gaussian_blur)

if __name__ == '__main__':
    napari.run()



"""
Add vectors image
=================

This example generates an image of vectors
Vector data is an array of shape (N, M, 2)
Each vector position is defined by an (x-proj, y-proj) element where
* x-proj and y-proj are the vector projections at each center
* each vector is centered on a pixel of the NxM grid

.. tags:: visualization-basic
"""

import numpy as np

import napari

# create the viewer and window
viewer = napari.Viewer()

n = 20
m = 40

image = 0.2 * np.random.random((n, m)) + 0.5
layer = viewer.add_image(image, contrast_limits=[0, 1], name='background')

# sample vector image-like data
# n x m grid of slanted lines
# random data on the open interval (-1, 1)
pos = np.zeros(shape=(n, m, 2), dtype=np.float32)
rand1 = 2 * (np.random.random_sample(n * m) - 0.5)
rand2 = 2 * (np.random.random_sample(n * m) - 0.5)

# assign projections for each vector
pos[:, :, 0] = rand1.reshape((n, m))
pos[:, :, 1] = rand2.reshape((n, m))

# add the vectors
vect = viewer.add_vectors(pos, edge_width=0.2, length=2.5)

print(image.shape, pos.shape)

if __name__ == '__main__':
    napari.run()



from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QCheckBox,
    QSlider,
    QWidget,
)

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import checked_to_bool, qt_signals_blocked
from napari.layers import Tracks
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtTailLengthSliderControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the current tail length
    attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Tracks
        An instance of a napari layer.

    Attributes
    ----------
    tail_length_slider : qtpy.QtWidgets.QSlider
        Slider controlling tail length of the layer.
    tail_length_slider_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the tail length chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Tracks) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.tail_length.connect(self._on_tail_length_change)

        # Setup widgets
        # slider for track tail length
        self.tail_length_slider = QSlider(Qt.Orientation.Horizontal)
        self.tail_length_slider.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.tail_length_slider.setMinimum(1)
        self.tail_length_slider.setMaximum(self._layer._max_length)
        self.tail_length_slider.setSingleStep(1)
        connect_setattr(
            self.tail_length_slider.valueChanged, self._layer, 'tail_length'
        )

        self.tail_length_slider_label = QtWrappedLabel(trans._('tail length:'))

        self._on_tail_length_change()

    def _on_tail_length_change(self) -> None:
        """Receive layer model track line width change event and update slider."""
        with qt_signals_blocked(self.tail_length_slider):
            value = self._layer.tail_length
            if value > self.tail_length_slider.maximum():
                self.tail_length_slider.setMaximum(self._layer._max_length)
            self.tail_length_slider.setValue(value)

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.tail_length_slider_label, self.tail_length_slider)]


class QtTailWidthSliderControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the current tail width
    attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Tracks
        An instance of a napari Tracks layer.

    Attributes
    ----------
    tail_width_slider : qtpy.QtWidgets.QSlider
        Slider controlling tail width of the layer.
    tail_width_slider_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the tail width chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Tracks) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.tail_width.connect(self._on_tail_width_change)
        self._layer.events.data.connect(self._on_data_change)

        # Setup widgets
        # slider for track edge width
        self.tail_width_slider = QSlider(Qt.Orientation.Horizontal)
        self.tail_width_slider.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.tail_width_slider.setMinimum(1)
        self.tail_width_slider.setMaximum(int(self._layer._max_width))
        self.tail_width_slider.setSingleStep(1)
        connect_setattr(
            self.tail_width_slider.valueChanged, self._layer, 'tail_width'
        )

        self.tail_width_slider_label = QtWrappedLabel(trans._('tail width:'))

        self._on_tail_width_change()

    def _on_tail_width_change(self) -> None:
        """Receive layer model track line width change event and update slider."""
        with qt_signals_blocked(self.tail_width_slider):
            value = int(self._layer.tail_width)
            self.tail_width_slider.setValue(value)

    def _on_data_change(self) -> None:
        self.tail_width_slider.setMaximum(int(self._layer._max_width))

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.tail_width_slider_label, self.tail_width_slider)]


class QtTailDisplayCheckBoxControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the tail should be
    displayed attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Tracks
        An instance of a napari Tracks layer.

    Attributes
    ----------
    tail_checkbox : qtpy.QtWidgets.QCheckBox
        Checkbox controlling if tails of the layer should be shown.
    tail_width_slider_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for showing the tails chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Tracks) -> None:
        super().__init__(parent, layer)
        # Setup layer
        # NOTE(arl): there are no events fired for changing checkbox (layer `display_tail` attribute)

        # Setup widgets
        self.tail_checkbox = QCheckBox()
        self.tail_checkbox.setChecked(True)
        connect_setattr(
            self.tail_checkbox.stateChanged,
            layer,
            'display_tail',
            convert_fun=checked_to_bool,
        )

        self.tail_checkbox_label = QtWrappedLabel(trans._('tail:'))

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.tail_checkbox_label, self.tail_checkbox)]



"""
Reader plugin
=============

Barebones reader plugin example, using ``imageio.imread```

.. tags:: historical
"""
from imageio import formats, imread
from napari_plugin_engine import napari_hook_implementation

readable_extensions = tuple({x for f in formats for x in f.extensions})


@napari_hook_implementation
def napari_get_reader(path):
    """A basic implementation of the napari_get_reader hook specification."""
    # if we know we cannot read the file, we immediately return None.
    if not path.endswith(readable_extensions):
        return None
    # otherwise we return the *function* that can read ``path``.
    return reader_function


def reader_function(path):
    """Take a path and returns a list of LayerData tuples."""
    data = imread(path)
    # Readers are expected to return data as a list of tuples, where each tuple
    # is (data, [meta_dict, [layer_type]])
    return [(data,)]




import numpy as np

import napari

# Set the number of steps
num_steps = 2**17

base = np.linspace(start=1, stop=num_steps, num=num_steps).astype('uint32')
label_img = np.repeat(
        base.reshape([1, base.shape[0]]), int(num_steps/1000), axis=0
        )

viewer = napari.Viewer()
viewer.add_image(
        label_img,
        scale=(100, 1),
        colormap='viridis',
        contrast_limits=(0, num_steps),
        )

if __name__ == '__main__':
    napari.run()



"""
magicgui dask delayed
=====================

An example of calling a threaded function from a magicgui dock_widget.
Note: this example requires python >= 3.9

.. tags:: gui
"""
import time
from concurrent.futures import Future

import dask.array as da
from magicgui import magicgui

import napari
from napari.types import ImageData


def _slow_function(nz):
    time.sleep(2)
    return da.random.random((nz, 512, 512))


if __name__ == '__main__':
    from dask.distributed import Client

    client = Client()

    @magicgui(client={'bind': client})
    def widget(client, nz: int = 1000) -> Future[ImageData]:
        return client.submit(_slow_function, nz)

    viewer = napari.Viewer()
    viewer.window.add_dock_widget(widget, area='right')

if __name__ == '__main__':
    napari.run()



from typing import Literal

from napari import __version__
from napari.components.overlays.base import CanvasOverlay


class WelcomeOverlay(CanvasOverlay):
    """Welcome screen overlay."""

    # not settable in this specific overlay
    position: None = None
    gridded: Literal[False] = False
    version: str = __version__
    shortcuts: tuple[str, ...] = (
        'napari.window.file._image_from_clipboard',
        'napari.window.file.open_files_dialog',
        'napari.window.view.toggle_command_palette',
        'napari:show_shortcuts',
    )
    # TODO: query mouse binding as well somehow? Currently we have to hardcode those.
    tips: tuple = (
        'You can take a screenshot of the canvas and copy it to your clipboard by pressing {napari.window.file.copy_canvas_screenshot}.',
        'You can change most shortcuts from the File → Preferences → Shortcuts menu.',
        'You can right click many components of the graphical interface to access advanced controls.',
        'If you select multiple layers in the layer list, then right click and select "Link Layers", their parameters will be synced.',
        'You can press {Ctrl} and scroll the mouse wheel to move the dimension sliders.',
        'To zoom in on a specific area, hold {Alt} and draw a rectangle around it.',
        'Hold {napari:hold_for_pan_zoom} to pan/zoom in any mode (e.g. while painting).',
        'While painting labels, hold {Alt} and move the cursor left/right to quickly decrease/increase the brush size.',
        'If you have questions, you can reach out on our community chat at napari.zulipchat.com!',
        'The community at forum.image.sc is full of imaging experts sharing knowledge and tools for napari and much, much more!',
    )



"""
Multiple viewers
================

Create multiple viewers from the same script

.. tags:: gui
"""

from skimage import data

import napari

# add the image
photographer = data.camera()
viewer_a = napari.Viewer(title="viewer a")
layer_a = viewer_a.add_image(photographer, name='photographer')

# add the image in a new viewer window
astronaut = data.astronaut()
# Also view_path, add_shapes, add_points, add_labels etc.
viewer_b = napari.Viewer(title="viewer b")
layer_b = viewer_b.add_image(astronaut, name='astronaut')

if __name__ == '__main__':
    napari.run()



from qtpy.QtWidgets import QWidget

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import attr_to_settr
from napari._qt.widgets.qt_color_swatch import QColorSwatchEdit
from napari.layers import Points
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtBorderColorControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the current border
    color layer attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Points
        An instance of a napari Points layer.

    Attributes
    ----------
    border_color_edit : napari._qt.widgets.qt_color_swatch.QColorSwatchEdit
        ColorSwatchEdit controlling current face color of the layer.
    border_color_edit_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the current egde color chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Points) -> None:
        super().__init__(parent, layer)
        # Setup widgets
        self.border_color_edit = QColorSwatchEdit(
            initial_color=self._layer.current_border_color,
            tooltip=trans._(
                'Click to set the border color of currently selected points and any added afterwards.'
            ),
        )
        connect_setattr(
            self.border_color_edit.color_changed,
            self._layer,
            'current_border_color',
        )
        self._callbacks.append(
            attr_to_settr(
                self._layer,
                'current_border_color',
                self.border_color_edit,
                'setColor',
            )
        )
        self._callbacks.append(
            attr_to_settr(
                self._layer._border,
                'current_color',
                self.border_color_edit,
                'setColor',
            )
        )

        self.border_color_edit_label = QtWrappedLabel(trans._('border color:'))

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.border_color_edit_label, self.border_color_edit)]



"""
magicgui with threadworker
==========================

An example of calling a threaded function from a magicgui ``dock_widget``.
Note: this example requires python >= 3.9

.. tags:: gui
"""
from typing import Annotated

from magicgui import magic_factory, widgets
from skimage import data
from skimage.feature import blob_log

import napari
from napari.qt.threading import FunctionWorker, thread_worker
from napari.types import ImageData, LayerDataTuple


@magic_factory(pbar={'visible': False, 'max': 0, 'label': 'working...'})
def make_widget(
    pbar: widgets.ProgressBar,
    image: ImageData,
    min_sigma: Annotated[float, {'min': 0.5, 'max': 15, 'step': 0.5}] = 5,
    max_sigma: Annotated[float, {'min': 1, 'max': 200, 'step': 0.5}] = 30,
    num_sigma: Annotated[int, {'min': 1, 'max': 20}] = 10,
    threshold: Annotated[float, {'min': 0, 'max': 1000, 'step': 0.1}] = 6,
) -> FunctionWorker[LayerDataTuple]:

    # @thread_worker creates a worker that runs a function in another thread
    # we connect the "returned" signal to the ProgressBar.hide method
    @thread_worker(connect={'returned': pbar.hide})
    def detect_blobs() -> LayerDataTuple:
        # this is the potentially long-running function
        blobs = blob_log(image, min_sigma, max_sigma, num_sigma, threshold)
        points = blobs[:, : image.ndim]
        meta = {
            'size': blobs[:, -1],
            'border_color': 'red',
            'border_width': 2,
            'border_width_is_relative': False,
            'face_color': 'transparent',
        }
        # return a "LayerDataTuple"
        return (points, meta, 'points')

    # show progress bar and return worker
    pbar.show()
    return detect_blobs()


viewer = napari.Viewer()
viewer.window.add_dock_widget(make_widget(), area='right')
viewer.add_image(data.hubble_deep_field().mean(-1))

napari.run()



"""
Launch napari and display an image
==================================

Launch an instance of napari and simultaneously display an image using the
:func:`imshow` API. This example can be run to recreate the README image.

.. tags:: visualization-basic
"""

from skimage import data

import napari

viewer, layers = napari.imshow(data.cells3d(), channel_axis=1, ndisplay=3)

# To recreate the README image uncomment the following lines:
# viewer.window._qt_window.resize(800, 600)
# napari._qt.qt_event_loop.get_qapp().processEvents()  # Ensure the window is done rendering before taking a screenshot
# viewer.screenshot('./src/napari/resources/multichannel_cells.png', canvas_only=False)

if __name__ == '__main__':
    napari.run()



"""
Add points with multicolor text
===============================

Display a points layer on top of an image layer with text using
multiple face colors mapped from features for the points and text.

.. tags:: visualization-basic
"""

import numpy as np

import napari

# add the image with three points
viewer = napari.Viewer()
layer = viewer.add_image(np.zeros((400, 400)))
points = np.array([[100, 100], [200, 300], [333, 111]])

# create features for each point
features = {
    'confidence': np.array([1, 0.5, 0]),
    'good_point': np.array([True, False, False]),
}

# define the color cycle for the points face and text colors
color_cycle = ['blue', 'green']

text = {
    'string': 'Confidence is {confidence:.2f}',
    'size': 20,
    'color': {'feature': 'good_point', 'colormap': color_cycle},
    'translation': np.array([-30, 0]),
}

# create a points layer where the face_color is set by the good_point feature
# and the border_color is set via a color map (grayscale) on the confidence
# feature
points_layer = viewer.add_points(
    points,
    features=features,
    text=text,
    size=20,
    border_width=7,
    border_width_is_relative=False,
    border_color='confidence',
    border_colormap='gray',
    face_color='good_point',
    face_color_cycle=color_cycle,
)

# set the border_color mode to colormap
points_layer.border_color_mode = 'colormap'

if __name__ == '__main__':
    napari.run()



"""
Ensure that layers and their convenience methods on the viewer
have the same signatures and docstrings.
"""

import gc
import inspect
import re

import numpy as np
import pytest
from docstring_parser.numpydoc import parse as parse_docstring

import napari
from napari import Viewer, layers as module
from napari._tests.utils import check_viewer_functioning
from napari.utils.misc import camel_to_snake

layers = []

for name in dir(module):
    obj = getattr(module, name)

    if obj is module.Layer or not inspect.isclass(obj):
        continue

    if issubclass(obj, module.Layer):
        layers.append(obj)


@pytest.mark.parametrize('layer', layers, ids=lambda layer: layer.__name__)
def test_docstring(layer):
    name = layer.__name__

    method_name = f'add_{camel_to_snake(name)}'
    method = getattr(Viewer, method_name)

    method_doc = parse_docstring(method.__doc__ or '')
    layer_doc = parse_docstring(layer.__doc__ or '')

    # check summary section
    method_summary = method_doc.short_description or ''

    if name == 'Image':
        summary_format = 'Add one or more Image layers to the layer list.'
    else:
        summary_format = 'Add an? .+? layers? to the layer list.'

    assert re.match(summary_format, method_summary), (
        f"improper 'Summary' section of '{method_name}'"
    )

    # check parameters section
    method_params = method_doc.params
    layer_params = layer_doc.params

    # Remove path parameter from viewer method if it exists
    method_params = [m for m in method_params if m.arg_name != 'path']

    if name == 'Image':
        # For Image just test arguments that are in layer are in method
        # Filter layer params to only those that are actually in the method
        # (layer docstrings may include attributes that aren't constructor params)
        method_param_names = [m.arg_name for m in method_params]
        layer_params = [
            p for p in layer_params if p.arg_name in method_param_names
        ]
        for layer_param in layer_params:
            assert layer_param.arg_name in method_param_names
    else:
        # For other layers, ensure all method params are documented in layer
        # (We don't check the reverse because layer docstrings may include
        # attributes that aren't constructor params)
        method_param_names = {m.arg_name for m in method_params}
        layer_param_dict = {p.arg_name: p for p in layer_params}

        # Check that all method params exist in layer docs
        for method_param in method_params:
            assert method_param.arg_name in layer_param_dict, (
                f"Parameter '{method_param.arg_name}' in {method_name} "
                f'not found in {name} docstring'
            )

        try:
            for method_param in method_params:
                layer_param = layer_param_dict[method_param.arg_name]

                m_name = method_param.arg_name
                l_name = layer_param.arg_name

                assert m_name == l_name, 'different parameter names'
                # Note: We don't check types or descriptions because they may
                # reasonably differ between the layer class and viewer method
                # perspectives (e.g., method may accept more types for convenience)
        except AssertionError as e:
            raise AssertionError(
                f"docstrings don't match for class {name}"
            ) from e

    # check returns section
    method_returns = method_doc.returns

    if method_returns:
        return_name = method_returns.return_name or 'layer'
        return_type = method_returns.type_name or ''
        return_description = (method_returns.description or '').strip()

        if name == 'Image':
            expected_type = f':class:`napari.layers.{name}` or list'
            expected_desc = f'The newly-created {name.lower()} layer or list of {name.lower()} layers.'
        else:
            expected_type = f':class:`napari.layers.{name}`'
            expected_desc = f'The newly-created {name.lower()} layer.'

        assert return_name == 'layer', (
            f"improper return name in '{method_name}'"
        )
        assert return_type == expected_type, (
            f"improper return type in '{method_name}'"
        )
        assert return_description == expected_desc, (
            f"improper return description in '{method_name}'"
        )


@pytest.mark.parametrize('layer', layers, ids=lambda layer: layer.__name__)
def test_signature(layer):
    name = layer.__name__

    method = getattr(Viewer, f'add_{camel_to_snake(name)}')

    class_parameters = dict(inspect.signature(layer.__init__).parameters)
    method_parameters = dict(inspect.signature(method).parameters)

    fail_msg = f"signatures don't match for class {name}"
    if name == 'Image':
        # If Image just test that class params appear in method
        for class_param in class_parameters:
            assert class_param in method_parameters, fail_msg
    else:
        assert class_parameters == method_parameters, fail_msg


# plugin_manager fixture is added to prevent errors due to installed plugins
def test_imshow(qtbot, napari_plugin_manager):
    shape = (10, 15)
    ndim = len(shape)
    np.random.seed(0)
    data = np.random.random(shape)
    viewer, layer = napari.imshow(data, channel_axis=None, show=False)
    view = viewer.window._qt_viewer
    check_viewer_functioning(viewer, view, data, ndim)
    assert isinstance(layer, napari.layers.Image)
    viewer.close()


# plugin_manager fixture is added to prevent errors due to installed plugins
def test_imshow_multichannel(qtbot, napari_plugin_manager):
    """Test adding image."""
    np.random.seed(0)
    data = np.random.random((15, 10, 5))
    viewer, layers = napari.imshow(data, channel_axis=-1, show=False)
    assert len(layers) == data.shape[-1]
    assert isinstance(layers, tuple)
    for i in range(data.shape[-1]):
        np.testing.assert_array_equal(layers[i].data, data.take(i, axis=-1))
    viewer.close()
    # Run a full garbage collection here so that any remaining viewer
    # and related instances are removed for future tests that may use
    # make_napari_viewer.
    gc.collect()


# plugin_manager fixture is added to prevent errors due to installed plugins
def test_imshow_with_viewer(qtbot, napari_plugin_manager, make_napari_viewer):
    shape = (10, 15)
    ndim = len(shape)
    np.random.seed(0)
    data = np.random.random(shape).astype(np.float32)
    viewer = make_napari_viewer()
    viewer2, layer = napari.imshow(data, viewer=viewer, show=False)
    assert viewer is viewer2
    np.testing.assert_array_equal(data, layer.data)
    view = viewer.window._qt_viewer
    check_viewer_functioning(viewer, view, data, ndim)
    viewer.close()



"""
nD shapes
=========

Display one 4-D image layer using the ``add_image`` API

.. tags:: visualization-nD
"""

import numpy as np
from skimage import data

import napari

blobs = data.binary_blobs(
    length=128, blob_size_fraction=0.05, n_dim=3, volume_fraction=0.1
).astype(float)

viewer = napari.Viewer()
layer = viewer.add_image(blobs.astype(float))

# create one random polygon per "plane"
planes = np.tile(np.arange(128).reshape((128, 1, 1)), (1, 5, 1))
np.random.seed(0)
corners = np.random.uniform(0, 128, size=(128, 5, 2))
shapes = np.concatenate((planes, corners), axis=2)

base_cols = ['red', 'green', 'blue', 'white', 'yellow', 'magenta', 'cyan']
colors = np.random.choice(base_cols, size=128)

layer = viewer.add_shapes(
    np.array(shapes),
    shape_type='polygon',
    face_color=colors,
    name='sliced',
)

masks = layer.to_masks(mask_shape=(128, 128, 128))
labels = layer.to_labels(labels_shape=(128, 128, 128))
shape_array = np.array(layer.data)

print(
    f'sliced: nshapes {layer.nshapes}, mask shape {masks.shape}, '
    f'labels_shape {labels.shape}, array_shape, {shape_array.shape}'
)

corners = np.random.uniform(0, 128, size=(2, 2))
layer = viewer.add_shapes(corners, shape_type='rectangle', name='broadcasted')

masks = layer.to_masks(mask_shape=(128, 128))
labels = layer.to_labels(labels_shape=(128, 128))
shape_array = np.array(layer.data)

print(
    f'broadcast: nshapes {layer.nshapes}, mask shape {masks.shape}, '
    f'labels_shape {labels.shape}, array_shape, {shape_array.shape}'
)

if __name__ == '__main__':
    napari.run()
